import datetime
from dataclasses import dataclass
from flask import request, current_app as app

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db, logger
from sqlalchemy import Integer, String, Boolean, JSON, DateTime, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.models.MissionRole import MissionRole


@dataclass
class Mission(db.Model):
    __tablename__ = "missions"

    INVITE = "INVITE"
    DELETE = "DELETE"
    CREATE = "CREATE"

    name: Mapped[str] = mapped_column(String(255), primary_key=True)
    description: Mapped[str] = mapped_column(String(255), nullable=True)
    chat_room: Mapped[str] = mapped_column(String(255), nullable=True)
    base_layer: Mapped[str] = mapped_column(String(255), nullable=True)
    bbox: Mapped[str] = mapped_column(String(255), nullable=True)
    path: Mapped[str] = mapped_column(String(255), nullable=True)
    classification: Mapped[str] = mapped_column(String(255), nullable=True)
    tool: Mapped[str] = mapped_column(String(255), nullable=True)
    group: Mapped[str] = mapped_column(String(255), nullable=True)
    default_role: Mapped[str] = mapped_column(String(255), nullable=True)
    keywords: Mapped[JSON] = mapped_column(JSON, nullable=True)
    creator_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True)
    create_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    external_data: Mapped[JSON] = mapped_column(JSON, nullable=True)
    feeds: Mapped[JSON] = mapped_column(JSON, nullable=True)
    map_layers: Mapped[JSON] = mapped_column(JSON, nullable=True)
    invite_only: Mapped[bool] = mapped_column(Boolean, nullable=True)
    expiration: Mapped[int] = mapped_column(Integer, nullable=False, default=-1)
    guid: Mapped[str] = mapped_column(String(255), nullable=True)
    password_protected: Mapped[bool] = mapped_column(Boolean, nullable=True)
    password: Mapped[str] = mapped_column(String(255), nullable=True)
    invitations = relationship("MissionInvitation", cascade="all, delete-orphan",  back_populates="mission")
    roles = relationship("MissionRole", cascade="all, delete-orphan",  back_populates="mission")
    mission_changes = relationship("MissionChange", cascade="all, delete-orphan",  back_populates="mission", uselist=True)
    contents = relationship("MissionContent", cascade="all, delete",  secondary="mission_content_mission", back_populates="mission", uselist=True)
    cots = relationship("CoT", back_populates="mission", uselist=True)
    uids = relationship("MissionUID", cascade="all, delete-orphan", back_populates="mission")
    mission_logs = relationship("MissionLogEntry", cascade="all, delete-orphan", back_populates="mission")
    owner = relationship("EUD", back_populates="owned_missions")
    groups = relationship("Group", secondary="groups_missions", viewonly=True, back_populates="missions", cascade="all, delete")

    def serialize(self):
        return {
            'description': self.description,
            'chat_room': self.chat_room,
            'base_layer': self.base_layer,
            'bbox': self.bbox,
            'path': self.path,
            'classification': self.classification,
            'tool': self.tool,
            'group': self.group or "__ANON__",
            'default_role': self.default_role,
            'keywords': self.keywords,
            'creator_uid': self.creator_uid,
            'create_time': self.create_time,
            'external_data': self.external_data,
            'feeds': self.feeds,
            'map_layers': self.map_layers,
            'invite_only': self.invite_only,
            'expiration': self.expiration,
            'guid': self.guid,
            'password_protected': self.password_protected,
            'password': self.password
        }

    def to_json(self):
        url = request.url_root.replace("http://", "").replace("https://", "").replace("/", "")
        json = {
            'name': self.name,
            'description': self.description or "",
            'chatRoom': self.chat_room or "",
            'baseLayer': self.base_layer or "",
            'bbox': self.bbox or "",
            'path': self.path or "",
            'classification': self.classification or "",
            'tool': self.tool or "public",
            'defaultRole': self.default_role or "",
            'keywords': self.keywords if self.keywords else [],
            'creatorUid': self.creator_uid or "",
            'createTime': iso8601_string_from_datetime(self.create_time) or "",
            'externalData': self.external_data if self.external_data else [],
            'feeds': self.feeds if self.feeds else [],
            'mapLayers': self.map_layers if self.map_layers else [],
            'inviteOnly': self.invite_only if self.invite_only is not None else False,
            'expiration': self.expiration if self.expiration is not None else -1,
            'guid': self.guid or "",
            'uids': [uid.to_json() for uid in self.uids],
            'contents': [content.to_json() for content in self.contents],
            'passwordProtected': self.password_protected if self.password_protected is not None else False,
            'missionChanges': [mission_change.to_json() for mission_change in self.mission_changes],
            'qr_code': f"{url}:{app.config.get('OTS_SSL_STREAMING_PORT')}:ssl,{url}-{app.config.get('OTS_MARTI_HTTPS_PORT')}-ssl-{self.name},{self.name}",
            'owner': self.owner.to_json() if self.owner else None,
            'groups': []
        }

        for group in self.groups:
            json['groups'].append({"id": group.id, "name": group.name})

        if self.default_role == MissionRole.MISSION_SUBSCRIBER or not self.default_role:
            json['defaultRole'] = MissionRole.SUBSCRIBER_ROLE

        elif self.default_role == MissionRole.MISSION_OWNER:
            json['defaultRole'] = MissionRole.OWNER_ROLE

        else:
            json['defaultRole'] = MissionRole.READ_ONLY_ROLE

        return json

    def to_marti_json(self):
        return_value = self.to_json()

        return_value['groups'] = []
        for group in self.groups:
            return_value['groups'].append(group.name)

        return return_value


import base64
import logging
import platform
from logging.handlers import TimedRotatingFileHandler

import colorlog
import datetime
import os
import time
import traceback
import sys
import random
import flask_sqlalchemy
from flask_socketio import SocketIO

import bleach
import pika
import sqlalchemy.exc
from flask_security import SQLAlchemyUserDatastore
from flask_security.models import fsqla
from pika.channel import Channel
from sqlalchemy import exc, insert, update, select
from bs4 import BeautifulSoup
from meshtastic import mqtt_pb2, mesh_pb2, portnums_pb2, BROADCAST_NUM
import unishox2
import yaml
from flask import Flask, jsonify

from opentakserver.defaultconfig import DefaultConfig
from opentakserver.functions import *
from opentakserver.models.EUDStats import EUDStats
from opentakserver.models.Meshtastic import MeshtasticChannel
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionLogEntry import MissionLogEntry
from opentakserver.models.MissionChange import MissionChange, generate_mission_change_cot
from opentakserver.models.GroupMission import GroupMission
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.WebAuthn import WebAuthn

from opentakserver.proto import atak_pb2
from opentakserver.extensions import logger, db
from opentakserver.functions import datetime_from_iso8601_string
from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.GeoChat import GeoChat
from opentakserver.models.Icon import Icon
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.ZMIST import ZMIST
from opentakserver.models.Point import Point
from opentakserver.models.Marker import Marker
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.Certificate import Certificate
from opentakserver.models.Group import Group


class CoTController:

    def __init__(self, context, log, database, socket_io):
        self.context = context
        self.logger = log
        self.db: flask_sqlalchemy.SQLAlchemy = database
        self.socketio = socket_io

        self.exchanges = []

        self.rabbit_connection: pika.BlockingConnection = None
        self.rabbit_channel = None

    def run(self):
        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        self.rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        self.rabbit_channel = self.rabbit_connection.channel()
        self.rabbit_channel.exchange_declare(exchange='cot', exchange_type='fanout', durable=True)
        self.rabbit_channel.queue_declare(queue='cot_parser')
        self.rabbit_channel.queue_bind(exchange='cot', queue='cot_parser')
        self.rabbit_channel.basic_qos(prefetch_count=self.context.app.config.get("OTS_RABBITMQ_PREFETCH"))
        self.rabbit_channel.basic_consume(queue='cot_parser', on_message_callback=self.on_message, auto_ack=False)
        self.rabbit_channel.start_consuming()

    def insert_cot(self, soup, event, uid):
        start = datetime_from_iso8601_string(event.attrs['start'])
        stale = datetime_from_iso8601_string(event.attrs['stale'])
        timestamp = datetime_from_iso8601_string(event.attrs['time'])

        # Assign CoT to a data sync mission
        dest = event.find("dest")
        mission_name = None
        if dest and 'mission' in dest.attrs:
            mission_name = dest.attrs['mission']

        with self.context:
            res = self.db.session.execute(insert(CoT).values(
                how=event.attrs['how'], type=event.attrs['type'], sender_uid=uid, timestamp=timestamp, xml=str(soup),
                start=start, stale=stale, mission_name=mission_name, uid=event.attrs['uid']
            ))

            try:
                self.db.session.commit()
                return res.inserted_primary_key[0]
            except sqlalchemy.exc.IntegrityError:
                # When using MySQL it will raise IntegrityError when a new EUD connects and it doesn't exist yet in the EUDs table
                # We'll ignore this error and not insert this CoT so the EUD table can be populated
                return None

    def parse_point(self, event, uid, cot_id):
        # hae = Height above the WGS ellipsoid in meters
        # ce = Circular 1-sigma or a circular area about the location in meters
        # le = Linear 1-sigma error or an altitude range about the location in meters
        point = event.find('point')
        if point and not point.attrs['lat'].startswith('999'):
            p = Point()
            p.uid = event.attrs['uid']
            p.device_uid = uid
            p.ce = point.attrs['ce']
            p.hae = point.attrs['hae']
            p.le = point.attrs['le']
            p.latitude = float(point.attrs['lat'])
            p.longitude = float(point.attrs['lon'])
            p.timestamp = datetime_from_iso8601_string(event.attrs['time'])
            p.cot_id = cot_id

            # We only really care about the rest of the data if there's a valid lat/lon
            if p.latitude == 0 and p.longitude == 0:
                return None

            track = event.find('track')
            if track:
                if 'course' in track.attrs and track.attrs['course'] != "9999999.0":
                    p.course = track.attrs['course']
                else:
                    p.course = 0

                if 'speed' in track.attrs and track.attrs['speed'] != "9999999.0":
                    p.speed = track.attrs['speed']
                else:
                    p.speed = 0

            # For TAK ICU and OpenTAK ICU CoT's with bearing from the compass
            sensor = event.find('sensor')
            if sensor:
                if 'azimuth' in sensor.attrs:
                    p.azimuth = sensor.attrs['azimuth']
                # Camera's field of view
                if 'fov' in sensor.attrs:
                    p.fov = sensor.attrs['fov']

            precision_location = event.find('precisionlocation')
            if precision_location and 'geolocationsrc' in precision_location.attrs:
                p.location_source = precision_location.attrs['geolocationsrc']
            elif precision_location and 'altsrc' in precision_location.attrs:
                p.location_source = precision_location.attrs['altsrc']
            elif event.attrs['how'] == 'm-g':
                p.location_source = 'GPS'

            status = event.find('status')
            if status:
                if 'battery' in status.attrs:
                    p.battery = status.attrs['battery']

            with (self.context):
                res = self.db.session.execute(insert(Point).values(
                    uid=p.uid, device_uid=p.device_uid, ce=p.ce, hae=p.hae, le=p.le, latitude=p.latitude,
                    longitude=p.longitude, timestamp=p.timestamp, cot_id=cot_id, location_source=p.location_source,
                    course=p.course, speed=p.speed, battery=p.battery, fov=p.fov, azimuth=p.azimuth)
                )

                # iTAK sucks. Instead of sending mission CoTs with a <dest mission="mission_name"> tag, it sends a normal CoT and
                # makes a POST to /Marti/api/missions/mission_name/contents. The POST happens faster than the CoT can be received and parsed,
                # so we're left with a row in the mission_uids table without most of the details that come from the CoT. Fortunately
                # the mission_uids.uid field corresponds to the CoT's event UID, so the row in mission_uids can be updated here.
                usericon = event.find('usericon')
                color = event.find('color')
                contact = event.find('contact')

                iconset_path = None
                if usericon and 'iconsetpath' in usericon.attrs:
                    iconset_path = usericon.attrs['iconsetpath']
                elif usericon and 'iconsetPath' in usericon.attrs:
                    iconset_path = usericon.attrs['iconsetPath']

                cot_color = None
                if color and 'argb' in color.attrs:
                    cot_color = color.attrs['argb']
                if color and 'value' in color.attrs:
                    cot_color = color.attrs['value']

                callsign = None
                if contact and 'callsign' in contact.attrs:
                    callsign = contact.attrs['callsign']

                self.db.session.execute(update(MissionUID).where(MissionUID.uid == event.attrs['uid']).values(
                    cot_type=event.attrs['type'], latitude=p.latitude, longitude=p.longitude, iconset_path=iconset_path,
                    color=cot_color, callsign=callsign
                ))

                self.db.session.commit()
                # Get the point from the DB with its related CoT
                p = self.db.session.execute(select(Point).filter(Point.id == res.inserted_primary_key[0])).first()[0]

                # This CoT is a position update for an EUD. Send it to socketio clients so it can be seen on the UI map
                # OpenTAK ICU position updates don't include the <takv> tag, but we still want to send the updated position
                # to the UI's map
                if event.find('takv') or event.find("__video"):
                    self.socketio.emit('point', p.to_json(), namespace='/socket.io')

                if self.context.app.config.get("OTS_ENABLE_MESHTASTIC"):
                    try:
                        eud = self.db.session.execute(select(EUD).filter_by(uid=uid)).first()

                        if not eud:
                            return res.inserted_primary_key[0]

                        eud = eud[0]

                        now = datetime.now(timezone.utc)
                        if eud.last_meshtastic_publish is None or (now - eud.last_meshtastic_publish.replace(tzinfo=timezone.utc)).total_seconds() >= self.context.app.config.get("OTS_MESHTASTIC_PUBLISH_INTERVAL"):

                            self.logger.debug("publishing position to mesh")
                            try:
                                eud.last_meshtastic_publish = now
                                self.db.session.execute(update(EUD).filter_by(uid=eud.uid).values(last_meshtastic_publish=now))
                                self.db.session.commit()

                                if eud.platform != "Meshtastic":
                                    mesh_user = mesh_pb2.User()
                                    setattr(mesh_user, "id", "!{:x}".format(eud.meshtastic_id))
                                    mesh_user.hw_model = mesh_pb2.HardwareModel.PRIVATE_HW
                                    mesh_user.short_name = p.device_uid[-4:]

                                    contact = event.find('contact')
                                    if contact:
                                        mesh_user.long_name = contact.attrs['callsign']

                                    # Rate limits how often to send NodeInfo messages
                                    if (now - eud.last_meshtastic_publish.replace(tzinfo=timezone.utc)).total_seconds() >= \
                                        self.context.app.config.get("OTS_MESHTASTIC_NODEINFO_INTERVAL") * self.context.app.config.get("OTS_MESHTASTIC_PUBLISH_INTERVAL"):

                                        # Note to future self: The Meshtastic firmware expects a User payload when the Portnum is NodeInfo
                                        # DO NOT SEND A NODEINFO PAYLOAD!
                                        encoded_message = mesh_pb2.Data()
                                        encoded_message.portnum = portnums_pb2.NODEINFO_APP
                                        encoded_message.payload = mesh_user.SerializeToString()
                                        self.publish_to_meshtastic(self.get_protobuf(encoded_message, uid=p.device_uid))

                                    position = mesh_pb2.Position()
                                    position.latitude_i = int(p.latitude / .0000001)
                                    position.longitude_i = int(p.longitude / .0000001)
                                    position.altitude = int(p.hae)
                                    position.timestamp = int(time.mktime(p.timestamp.timetuple()))
                                    position.ground_track = int(p.course) if p.course else 0
                                    position.ground_speed = int(p.speed) if p.speed and p.speed >= 0 else 0
                                    position.seq_number = 1
                                    position.precision_bits = 32

                                    encoded_message = mesh_pb2.Data()
                                    encoded_message.portnum = portnums_pb2.POSITION_APP
                                    encoded_message.payload = position.SerializeToString()
                                    self.publish_to_meshtastic(self.get_protobuf(encoded_message, uid=p.device_uid))

                                    tak_packet = atak_pb2.TAKPacket()
                                    tak_packet.is_compressed = True
                                    tak_packet.contact.device_callsign, size = unishox2.compress(eud.uid)
                                    tak_packet.contact.callsign, size = unishox2.compress(eud.callsign)
                                    tak_packet.group.team = eud.team.name.replace(" ", "_") if eud.team else "Cyan"
                                    tak_packet.group.role = eud.team_role.replace(" ", "") if eud.team_role else "TeamMember"
                                    tak_packet.status.battery = int(p.battery) if p.battery else 0
                                    tak_packet.pli.latitude_i = int(p.latitude / .0000001)
                                    tak_packet.pli.longitude_i = int(p.longitude / .0000001)
                                    tak_packet.pli.altitude = int(p.hae) if p.hae else 0
                                    tak_packet.pli.speed = int(p.speed) if p.speed else 0
                                    tak_packet.pli.course = int(p.course) if p.course else 0

                                    encoded_message = mesh_pb2.Data()
                                    encoded_message.portnum = portnums_pb2.ATAK_PLUGIN
                                    encoded_message.payload = tak_packet.SerializeToString()

                                    self.publish_to_meshtastic(self.get_protobuf(encoded_message, uid=eud.uid, from_id=eud.meshtastic_id))
                            except BaseException as e:
                                self.logger.error(f"Failed to send publish message to mesh: {e}")
                                self.logger.debug(traceback.format_exc())

                    except BaseException as e:
                        logger.warning(f"Failed to publish Meshtastic message: {e}")
                        logger.debug(traceback.format_exc())

                return res.inserted_primary_key[0]

    def get_protobuf(self, payload, uid=None, from_id=None, to_id=BROADCAST_NUM, channel_id="LongFast"):
        if uid and not from_id:
            try:
                eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()[0]
                from_id = eud.meshtastic_id
            except:
                self.logger.error("Failed to find EUD {}, using random Meshtastic ID".format(uid))
                self.logger.debug(traceback.format_exc())
                from_id = random.getrandbits(32)

        message_id = random.getrandbits(32)
        mesh_packet = mesh_pb2.MeshPacket()
        mesh_packet.id = message_id

        try:
            setattr(mesh_packet, "from", int(from_id, 16))
        except BaseException as e:
            setattr(mesh_packet, "from", from_id)
        mesh_packet.to = to_id
        mesh_packet.want_ack = False
        mesh_packet.hop_limit = 3
        mesh_packet.decoded.CopyFrom(payload)

        service_envelope = mqtt_pb2.ServiceEnvelope()
        service_envelope.packet.CopyFrom(mesh_packet)
        service_envelope.channel_id = channel_id
        service_envelope.gateway_id = "OpenTAKServer"

        return service_envelope

    def parse_geochat(self, event, cot_id, point_pk):
        chat = event.find('__chat')
        if chat:
            chat_group = event.find('chatgrp')
            remarks = event.find('remarks')

            # Sometimes WinTAK seems to send GeoChat CoTs without remarks
            if not remarks:
                return

            chatroom = Chatroom()

            chatroom.name = chat.attrs['chatroom']
            chatroom.id = chat.attrs['id']
            chatroom.parent = chat.attrs['parent'] if 'parent' in chat.attrs else None

            with self.context:
                try:
                    self.db.session.add(chatroom)
                    self.db.session.commit()
                except exc.IntegrityError:
                    self.db.session.rollback()

            geochat = GeoChat()

            geochat.uid = event.attrs['uid']
            geochat.chatroom_id = chat.attrs['id']
            geochat.sender_uid = chat_group.attrs['uid0']
            geochat.remarks = remarks.text
            geochat.timestamp = datetime_from_iso8601_string(remarks.attrs['time'])
            geochat.point_id = point_pk
            geochat.cot_id = cot_id

            if self.context.app.config.get("OTS_ENABLE_MESHTASTIC"):
                try:
                    with self.context:
                        from_eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=geochat.sender_uid)).first()[0]

                        tak_packet = atak_pb2.TAKPacket()
                        tak_packet.contact.device_callsign, size = unishox2.compress(geochat.sender_uid)
                        tak_packet.contact.callsign, size = unishox2.compress(from_eud.callsign)
                        tak_packet.chat.message, size = unishox2.compress(remarks.text)
                        tak_packet.group.team = from_eud.team.name.replace(" ", "_")
                        tak_packet.group.role = from_eud.team_role.replace(" ", "")
                        tak_packet.is_compressed = True

                        send_meshtastic_text = False
                        dest = event.find("dest")
                        if dest and 'callsign' in dest.attrs and dest.attrs['callsign'] == chat.attrs['chatroom']:
                            # This is a DM
                            to = chat.attrs['id']
                            try:
                                # DM to a Meshtastic device
                                to_id = int(to, 16)
                                send_meshtastic_text = True
                            except:
                                # DM to an EUD running the Meshtastic ATAK Plugin
                                to_id = BROADCAST_NUM

                            tak_packet.chat.to, size = unishox2.compress(to)
                        else:
                            # This goes to a chat room
                            to = chat.attrs['chatroom']
                            to_id = BROADCAST_NUM
                            tak_packet.chat.to, size = unishox2.compress(to)

                            # By only sending a Meshtastic Data packet and not a TAK Packet, both the Meshtastic app
                            # and the Meshtastic ATAK plugin will receive the message
                            send_meshtastic_text = (to == "All Chat Rooms")

                        if send_meshtastic_text:
                            self.logger.debug("Publishing text to Meshtastic devices")
                            # Publish again for Meshtastic devices without the ATAK Plugin
                            encoded_message = mesh_pb2.Data()
                            encoded_message.portnum = portnums_pb2.TEXT_MESSAGE_APP
                            encoded_message.payload = remarks.text.encode("utf-8")
                            self.publish_to_meshtastic(
                                self.get_protobuf(encoded_message, to_id=to_id, from_id=from_eud.meshtastic_id))
                        else:
                            # Publish once for EUDs using the Meshtastic ATAK Plugin
                            encoded_message = mesh_pb2.Data()
                            encoded_message.portnum = portnums_pb2.ATAK_PLUGIN
                            encoded_message.payload = tak_packet.SerializeToString()

                            self.publish_to_meshtastic(
                                self.get_protobuf(encoded_message, from_id=from_eud.meshtastic_id, to_id=to_id))

                except BaseException as e:
                    self.logger.error("Failed to publish MQTT message: {}".format(e))
                    self.logger.debug(traceback.format_exc())

            with self.context:
                try:

                    self.db.session.add(geochat)
                    self.db.session.commit()
                except exc.IntegrityError:
                    # TODO: Check if remarks can be edited and if so do an update here
                    self.db.session.rollback()

            for attr in chat_group.attrs:
                if attr.startswith("uid") and attr != 'uid':

                    if 'groupOwner' in chat.attrs and chat.attrs['groupOwner'].lower() == 'true' and attr == 'uid0':
                        with self.context:
                            self.db.session.execute(update(Chatroom).where(Chatroom.id == chat.attrs['id'])
                                                    .values(group_owner=chat_group.attrs[attr]))
                            self.db.session.commit()

                    chatroom_uid = ChatroomsUids()
                    chatroom_uid.chatroom_id = chat.attrs['id']
                    chatroom_uid.uid = chat_group.attrs[attr]

                    with self.context:
                        try:
                            self.db.session.add(chatroom_uid)
                            self.db.session.commit()
                            self.logger.debug(
                                "add {} to chatroom {}".format(chatroom_uid.uid, chatroom_uid.chatroom_id))
                        except exc.IntegrityError:
                            self.db.session.rollback()

    def parse_video(self, event, cot_pk):
        video = event.find("__video")
        if video:
            self.logger.debug("Got video stream")
            connection_entry = video.find('ConnectionEntry')
            if not connection_entry:
                return

            path = connection_entry.attrs['path']
            if path.startswith("/"):
                path = path[1:]

            v = VideoStream()
            v.network_timeout = connection_entry.attrs['networkTimeout']
            v.uid = connection_entry.attrs['uid']
            v.path = path
            v.protocol = connection_entry.attrs['protocol']
            v.buffer_time = connection_entry.attrs['bufferTime']
            v.port = connection_entry.attrs['port']
            v.rover_port = connection_entry.attrs['roverPort']
            v.rtsp_reliable = connection_entry.attrs['rtspReliable']
            v.ignore_embedded_klv = (connection_entry.attrs['ignoreEmbeddedKLV'].lower() == 'true')
            v.alias = connection_entry.attrs['alias']
            v.cot_id = cot_pk
            v.generate_xml(connection_entry.attrs['address'])

            with self.context:
                try:
                    self.db.session.add(v)
                    self.db.session.commit()
                    self.logger.debug("Added video")
                except exc.IntegrityError as e:
                    self.db.session.rollback()
                    self.db.session.execute(update(VideoStream).where(VideoStream.uid == connection_entry.attrs['uid'])
                                            .values(network_timeout=connection_entry.attrs['networkTimeout'],
                                                    protocol=connection_entry.attrs['protocol'],
                                                    buffer_time=connection_entry.attrs['bufferTime'],
                                                    # address=connection_entry.attrs['address'],
                                                    port=connection_entry.attrs['port'],
                                                    rover_port=connection_entry.attrs['roverPort'],
                                                    rtsp_reliable=connection_entry.attrs['rtspReliable'],
                                                    ignore_embedded_klv=(connection_entry.attrs[
                                                                             'ignoreEmbeddedKLV'].lower() == 'true'),
                                                    alias=connection_entry.attrs['alias'],
                                                    xml=v.xml))

                    self.db.session.commit()

    def parse_alert(self, event, uid, point_pk, cot_pk):
        emergency = event.find('emergency')
        if emergency:
            if 'type' in emergency.attrs:
                emergency_type = emergency.attrs['type']
                alert = Alert()
                alert.sender_uid = uid
                alert.uid = event.attrs['uid']
                alert.start_time = datetime_from_iso8601_string(event.attrs['start'])
                alert.point_id = point_pk
                alert.alert_type = emergency_type
                alert.cot_id = cot_pk

                with self.context:
                    self.db.session.add(alert)
                    self.db.session.commit()
                    self.socketio.emit('alert', alert.to_json(), namespace='/socket.io')
            elif 'cancel' in emergency.attrs:
                with self.context:
                    try:
                        alert = self.db.session.execute(
                            Alert.query.filter(Alert.cancel_time == None, Alert.sender_uid == uid).order_by(
                                Alert.start_time.desc())).first()[0]
                        alert.cancel_time = datetime_from_iso8601_string(event.attrs['start'])
                        self.db.session.commit()
                        self.socketio.emit('alert', alert.to_json(), namespace='/socket.io')
                    except BaseException as e:
                        self.logger.error("Failed to set alert cancel time: {}".format(e))
                        self.logger.debug(traceback.format_exc())

    def parse_casevac(self, event, uid, point_pk, cot_pk):
        medevac = event.find('_medevac_')
        if medevac:
            zmist = medevac.find('zMist')
            with self.context:
                for a in medevac.attrs:
                    if medevac.attrs[a].lower() == 'true':
                        medevac.attrs[a] = True
                    elif medevac.attrs[a].lower() == 'false':
                        medevac.attrs[a] = False

                try:
                    self.db.session.execute(
                        insert(CasEvac).values(timestamp=datetime_from_iso8601_string(event.attrs['start']),
                                               sender_uid=uid, uid=event.attrs['uid'],
                                               point_id=point_pk, cot_id=cot_pk,
                                               **medevac.attrs))

                    if zmist:
                        self.db.session.execute(insert(ZMIST).values(casevac_uid=event.attrs['uid'], **zmist.attrs))
                except exc.IntegrityError as e:
                    self.db.session.rollback()
                    self.db.session.execute(update(CasEvac).where(CasEvac.uid == event.attrs['uid'])
                                            .values(**medevac.attrs))

                    if zmist:
                        self.db.session.execute(
                            update(ZMIST).where(CasEvac.uid == event.attrs['uid']).values(**zmist.attrs))
                self.db.session.commit()

                try:
                    casevac: CasEvac = \
                    self.db.session.execute(self.db.session.query(CasEvac).filter_by(uid=event.attrs['uid'])).first()[0]
                    self.socketio.emit('casevac', casevac.to_json(), namespace='/socket.io')
                except BaseException as e:
                    self.logger.error(f"Failed to emit CasEvac: {e}")
                    self.logger.debug(traceback.format_exc())

    def parse_marker(self, event, uid, point_pk, cot_pk):
        if ((re.match("^a-[f|h|u|p|a|n|s|j|k]-[Z|P|A|G|S|U|F]", event.attrs['type']) or
             # Spot map
             re.match("^b-m-p", event.attrs['type'])) and
                # Don't worry about EUD location updates
                not event.find('takv') and
                # Ignore video streams from sources like OpenTAK ICU
                event.attrs['type'] != 'b-m-p-s-p-loc'):

            try:
                marker = Marker()
                marker.uid = event.attrs['uid']
                marker.affiliation = get_affiliation(event.attrs['type'])
                marker.battle_dimension = get_battle_dimension(event.attrs['type'])
                marker.mil_std_2525c = cot_type_to_2525c(event.attrs['type'])

                detail = event.find('detail')
                icon = None

                if detail:
                    for tag in detail.find_all():
                        if 'readiness' in tag.attrs:
                            marker.readiness = tag.attrs['readiness'] == "true"
                        if 'argb' in tag.attrs:
                            marker.argb = tag.attrs['argb']
                            marker.color_hex = marker.color_to_hex()
                        if tag.name == "contact":
                            marker.callsign = tag.attrs['callsign']
                        if 'iconsetpath' in tag.attrs:
                            marker.iconset_path = tag.attrs['iconsetpath']
                            if marker.iconset_path.lower().endswith('.png'):
                                with self.context:
                                    filename = marker.iconset_path.split("/")[-1]

                                    try:
                                        icon = self.db.session.execute(
                                            self.db.session.query(Icon).filter(Icon.filename == filename)).first()[0]
                                        marker.icon_id = icon.id
                                    except:
                                        icon = self.db.session.execute(self.db.session.query(Icon).filter(
                                            Icon.filename == 'marker-icon.png')).first()
                                        if icon is None:
                                            marker.icon_id = None
                                        else:
                                            marker.icon_id = icon.id
                            elif not marker.mil_std_2525c:
                                with self.context:
                                    icon = self.db.session.execute(self.db.session.query(Icon)
                                                                   .filter(Icon.filename == 'marker-icon.png')).first()[
                                        0]
                                    marker.icon_id = icon.id

                        if 'altsrc' in tag.attrs:
                            marker.location_source = tag.attrs['altsrc']

                link = event.find('link')
                if link:
                    marker.parent_callsign = link.attrs['parent_callsign'] if 'parent_callsign' in link.attrs else None
                    marker.production_time = link.attrs[
                        'production_time'] if 'production_time' in link.attrs else iso8601_string_from_datetime(
                        datetime.now(timezone.utc))
                    marker.relation = link.attrs['relation'] if 'relation' in link.attrs else None
                    marker.relation_type = link.attrs['relation_type'] if 'relation_type' in link.attrs else None
                    marker.parent_uid = link.attrs['uid'] if 'uid' in link.attrs else None
                else:
                    marker.production_time = iso8601_string_from_datetime(datetime.now(timezone.utc))

                marker.point_id = point_pk
                marker.cot_id = cot_pk

                with self.context:
                    try:
                        self.db.session.add(marker)
                        self.db.session.commit()
                        self.logger.debug('added marker')
                    except exc.IntegrityError:
                        self.db.session.rollback()
                        self.db.session.execute(
                            update(Marker).where(Marker.uid == marker.uid).values(point_id=marker.point_id,
                                                                                  icon_id=marker.icon_id,
                                                                                  **marker.serialize()))
                        self.db.session.commit()
                        self.logger.debug('updated marker')
                        marker = self.db.session.execute(self.db.session.query(Marker)
                                                         .filter(Marker.uid == marker.uid)).first()[0]

                    self.socketio.emit('marker', marker.to_json(), namespace='/socket.io')

            except BaseException as e:
                self.logger.error("Failed to parse marker: {}".format(e))
                self.logger.debug(traceback.format_exc())

    def parse_rbline(self, event, uid, point_pk, cot_pk):
        if re.match("^u-rb", event.attrs['type']):
            self.logger.debug("Got an R&B line")
            rb_line = RBLine()

            detail = event.find('detail')
            if detail:
                rb_line.uid = event.attrs['uid']
                rb_line.sender_uid = uid
                rb_line.timestamp = datetime_from_iso8601_string(event.attrs['start'])
                rb_line.point_id = point_pk
                rb_line.cot_id = cot_pk

                for tag in detail:
                    if tag.name == 'range':
                        rb_line.range = tag.attrs['value']
                    if tag.name == 'bearing':
                        rb_line.bearing = tag.attrs['value']
                    # Sometimes ATAK sends NaN for the inclination which causes issues in the DB
                    if tag.name == 'inclination' and tag.attrs['value'].isnumeric():
                        rb_line.inclination = tag.attrs['value']
                    if tag.name == 'anchorUID':
                        rb_line.anchor_uid = tag.attrs['value']
                    if tag.name == 'rangeUnits':
                        rb_line.range_units = tag.attrs['value']
                    if tag.name == 'bearingUnits':
                        rb_line.bearing_units = tag.attrs['value']
                    if tag.name == 'northRef':
                        rb_line.north_ref = tag.attrs['value']
                    if tag.name == 'color':
                        rb_line.color = tag.attrs['value']
                        rb_line.color_hex = rb_line.color_to_hex()
                    if tag.name == 'contact':
                        rb_line.callsign = tag.attrs['callsign']
                    if tag.name == 'strokeColor':
                        rb_line.stroke_color = tag.attrs['value']
                    if tag.name == 'strokeWeight':
                        rb_line.stroke_weight = tag.attrs['value']
                    if tag.name == 'labels_on':
                        rb_line.labels_on = (tag.attrs['value'] == 'true')

                with self.context:

                    start_point = \
                        self.db.session.execute(self.db.session.query(Point).filter(Point.id == point_pk)).first()[0]
                    end_point = rb_line.calc_end_point(start_point)
                    rb_line.end_latitude = end_point['latitude']
                    rb_line.end_longitude = end_point['longitude']

                    try:
                        self.db.session.add(rb_line)
                        self.db.session.commit()
                        self.logger.debug("Inserted new R&B line: {}".format(rb_line.uid))
                    except exc.IntegrityError:
                        self.db.session.rollback()
                        self.db.session.execute(update(RBLine).where(RBLine.uid == rb_line.uid)
                                                .values(**rb_line.serialize()))
                        self.db.session.commit()
                        self.logger.debug('Updated R&B line: {}'.format(rb_line.uid))

                    rb_line.point = start_point
                    self.socketio.emit("rb_line", rb_line.to_json(), namespace='/socket.io')

    def parse_stats(self, event, uid):
        stats = event.find('stats')
        eud_stats = EUDStats()
        if stats:
            eud_stats.timestamp = datetime_from_iso8601_string(event.attrs.get('time'))
            eud_stats.eud_uid = uid
            eud_stats.battery_status = stats.attrs.get('battery_status')
            eud_stats.ip_address = stats.attrs.get('ip_address')
            if stats.attrs.get('app_framerate'):
                eud_stats.app_framerate = int(stats.attrs.get('app_framerate'))
            if stats.attrs.get('deviceDataRx'):
                eud_stats.deviceDataRx = int(stats.attrs.get('deviceDataRx'))
            if stats.attrs.get('deviceDataTx'):
                eud_stats.deviceDataTx = int(stats.attrs.get('deviceDataTx'))
            if stats.attrs.get('heap_current_size'):
                eud_stats.heap_current_size = int(stats.attrs.get('heap_current_size'))
            if stats.attrs.get('heap_free_size'):
                eud_stats.heap_free_size = int(stats.attrs.get('heap_free_size'))
            if stats.attrs.get('heap_max_size'):
                eud_stats.heap_max_size = int(stats.attrs.get('heap_max_size'))
            if stats.attrs.get('storage_available'):
                eud_stats.storage_available = int(stats.attrs.get('storage_available'))
            if stats.attrs.get('storage_total'):
                eud_stats.storage_total = int(stats.attrs.get('storage_total'))
            if stats.attrs.get('battery_temp'):
                eud_stats.battery_temp = int(stats.attrs.get('battery_temp'))
            if stats.attrs.get('battery'):
                eud_stats.battery = int(stats.attrs.get('battery').replace('%', ''))

            with self.context:
                self.db.session.add(eud_stats)
                self.db.session.commit()

    def publish_to_meshtastic(self, body):
        for channel in self.get_meshtastic_channels():
            body.channel_id = channel
            routing_key = "{}.2.e.{}.outgoing".format(self.context.app.config.get("OTS_MESHTASTIC_TOPIC"), channel)
            self.rabbit_channel.basic_publish(exchange='amq.topic', routing_key=routing_key,
                                              body=body.SerializeToString(),
                                              properties=pika.BasicProperties(
                                                  expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")))
            self.logger.debug("Published message to " + routing_key)

    def get_meshtastic_channels(self):
        with self.context:
            channels = self.db.session.execute(self.db.session.query(MeshtasticChannel)).scalars()
            downlink_channels = []
            for channel in channels:
                if channel.downlink_enabled:
                    downlink_channels.append(channel.name)
            return downlink_channels

    def on_message(self, channel: pika.channel.Channel, basic_deliver: pika.spec.Basic.Deliver, properties: pika.spec.BasicProperties, body: bytes):
        try:
            start = time.time()
            body = json.loads(body)
            soup = BeautifulSoup(body['cot'], 'xml')
            event: BeautifulSoup = soup.find('event')

            uid = body['uid'] or event.attrs['uid']
            if uid == self.context.app.config['OTS_NODE_ID']:
                uid = None

            if event:
                cot_pk = self.insert_cot(soup, event, uid)
                point_pk = self.parse_point(event, uid, cot_pk)
                self.parse_geochat(event, cot_pk, point_pk)
                self.parse_video(event, cot_pk)
                self.parse_alert(event, uid, point_pk, cot_pk)
                self.parse_casevac(event, uid, point_pk, cot_pk)
                self.parse_marker(event, uid, point_pk, cot_pk)
                self.parse_rbline(event, uid, point_pk, cot_pk)
                self.parse_stats(event, uid)
                self.rabbit_channel.basic_ack(delivery_tag=basic_deliver.delivery_tag)

                # EUD went offline
                if event.attrs['type'] == 't-x-d-d':

                    try:
                        with self.context:
                            eud = self.db.session.execute(self.db.session.query(EUD).filter_by(uid=uid)).first()
                            if eud:
                                eud = eud[0]
                                eud.last_event_time = datetime_from_iso8601_string(event.attrs['start'])
                                eud.last_status = 'Disconnected'
                                self.db.session.commit()
                                self.logger.debug("Updated {}".format(uid))
                                eud_json = eud.to_json()
                                # The first time an EUD connects but doesn't have a location.
                                # Tells the UI what kind of EUD this is, ie ATAK/WinTAK/iTAK or OpenTAK ICU
                                if not eud_json['last_point']:
                                    eud_json['type'] = event.attrs['type']
                                self.socketio.emit('eud', eud.to_json(), namespace='/socket.io')
                    except BaseException as e:
                        self.logger.error("Failed to update EUD: {}".format(e))
                        self.logger.debug(traceback.format_exc())

        except BaseException as e:
            self.logger.error(f"Failed to parse CoT: {e}")
            self.logger.debug(traceback.format_exc())
            self.rabbit_channel.basic_nack(delivery_tag=basic_deliver.delivery_tag, requeue=False)


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG
    logger.setLevel(level)

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            '%(log_color)s[%(asctime)s] - cot_parser[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s', datefmt="%Y-%m-%d %H:%M:%S")
        color_log_handler.setFormatter(color_log_formatter)
        logger.addHandler(color_log_handler)
        logger.info("Added color logger")

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'logs', 'opentakserver.log'),
                                  when=app.config.get("OTS_LOG_ROTATE_WHEN"), interval=app.config.get("OTS_LOG_ROTATE_INTERVAL"),
                                  backupCount=app.config.get("OTS_BACKUP_COUNT"))
    fh.setFormatter(logging.Formatter("[%(asctime)s] - cot_parser[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s"))
    logger.addHandler(fh)


def create_app():
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)

    # Load config.yml if it exists
    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
        app.config.from_file(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load)
    else:
        # First run, created config.yml based on default settings
        logger.info("Creating config.yml")
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
            conf = {}
            for option in DefaultConfig.__dict__:
                # Fix the sqlite DB path on Windows
                if option == "SQLALCHEMY_DATABASE_URI" and platform.system() == "Windows" and DefaultConfig.__dict__[option].startswith("sqlite"):
                    conf[option] = DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                elif option.isupper():
                    conf[option] = DefaultConfig.__dict__[option]
            config.write(yaml.safe_dump(conf))

    setup_logging(app)
    db.init_app(app)

    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.user import User
    from opentakserver.models.role import Role

    user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)

    return app


app = create_app()
child_processes = []

@app.route("/status")
def status():
    return jsonify({"status": "ok"})


def main():
    sio = SocketIO(message_queue="amqp://" + app.config.get("OTS_RABBITMQ_SERVER_ADDRESS"))

    processes = 0
    while processes < app.config.get("OTS_COT_PARSER_PROCESSES"):
        try:
            pid = os.fork()
            if pid == 0:
                cot_parser = CoTController(app.app_context(), logger, db, sio)
                cot_parser.run()
            else:
                child_processes.append(pid)
        except KeyboardInterrupt:
            pass
        except BaseException as e:
            logger.error(f"cot_parser error: {e}")
            logger.debug(traceback.format_exc())
        processes += 1

    for i, child in enumerate(child_processes):
        try:
            os.waitpid(child, 0)
        except BaseException:
            logger.info(f"Exiting...")
            sys.exit()


if __name__ == "__main__":
    main()



import os
import traceback

import pathlib

import bleach
from flask import current_app as app, request, Blueprint, jsonify, send_from_directory
from flask_babel import gettext
from flask_security import auth_required

from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.extensions import logger, db

from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.VideoRecording import VideoRecording

video_api_blueprint = Blueprint('video_api_blueprint', __name__)


@video_api_blueprint.route('/api/videos/thumbnail', methods=['GET'])
@auth_required()
def thumbnail():
    path = request.args.get("path")
    recording = request.args.get("recording")
    if not path:
        return jsonify({"success": False, "error": "Please specify a path"}), 400

    if recording and os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path, recording + ".png")):
        return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path),
                                   recording + ".png")

    elif os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path)):
        return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path),
                                   "thumbnail.png")

    return jsonify({"success": False, "error": gettext(u"Please specify a valid path")}), 400


@video_api_blueprint.route('/api/videos/recordings')
@auth_required()
def video_recordings():
    query = db.session.query(VideoRecording)
    query = search(query, VideoRecording, 'path')

    return paginate(query)


@video_api_blueprint.route('/api/videos/recording', methods=['GET', 'DELETE', 'HEAD'])
@auth_required()
def download_recording():
    if not request.args.get("id"):
        return jsonify({'success': False, 'error': 'Please specify a recording ID'}), 400
    recording_id = bleach.clean(request.args.get('id'))

    if recording_id.isdigit():
        recording_id = int(recording_id)
    else:
        return jsonify({"success": False, "error": gettext(u"Invalid recording_id: %(recording_id)s", recording_id=recording_id)}), 400

    try:
        recording = \
            db.session.execute(db.session.query(VideoRecording).filter(VideoRecording.id == recording_id)).first()[0]

        if request.method == 'GET':
            filename = pathlib.Path(recording.segment_path)
            return send_from_directory(filename.parent, filename.name)
        elif request.method == 'DELETE':
            os.remove(recording.segment_path)
            db.session.delete(recording)
            db.session.commit()
            return jsonify({'success': True})
        elif request.method == 'HEAD':
            return '', 200
    except:
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u'Recording not found')}), 404


@video_api_blueprint.route('/api/video_streams')
@auth_required()
def get_video_streams():
    query = db.session.query(VideoStream)
    query = search(query, VideoStream, 'username')
    query = search(query, VideoStream, 'protocol')
    query = search(query, VideoStream, 'path')
    query = search(query, VideoStream, 'uid')

    return paginate(query)



from wtforms import StringField, IntegerField, BooleanField, DateTimeField
from wtforms.validators import DataRequired, UUID

from opentakserver.forms.point_form import PointForm


class MarkerForm(PointForm):
    uid = StringField(validators=[DataRequired(), UUID()])
    callsign = StringField(validators=[DataRequired()])
    affiliation = StringField()
    battle_dimension = StringField()
    readiness = BooleanField()
    argb = IntegerField()
    color_hex = StringField()
    iconset_path = StringField()
    parent_callsign = StringField()
    production_time = DateTimeField(format="%Y-%m-%dT%H:%M:%S.%fZ")
    relation = StringField()
    relation_type = StringField()
    location_source = StringField()
    remarks = StringField()
    mil_std_2525c = StringField()



import base64
import os
import traceback

import sqlalchemy
from flask_babel import gettext
from meshtastic import channel_pb2, apponly_pb2

import bleach
from flask import Blueprint, request, jsonify, current_app as app
from flask_security import auth_required

from opentakserver.extensions import logger, db
from opentakserver.models.Meshtastic import MeshtasticChannel
from opentakserver.blueprints.ots_api.api import paginate, search

meshtastic_api_blueprint = Blueprint("meshtastic_api_blueprint", __name__)


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=['POST'])
@auth_required()
def create_channel():
    channel_set = apponly_pb2.ChannelSet()
    channel_settings = channel_pb2.ChannelSettings()

    # Parse the settings from a Meshtastic URL
    if 'url' in request.json.keys() and request.json.get('url'):
        try:
            settings = base64.urlsafe_b64decode(bleach.clean(request.json.get('url').split("#")[-1]) + "==")
            channel_set.ParseFromString(settings)
            url = request.json.get('url')

        except BaseException as e:
            logger.error("Failed to parse Meshtastic  URL: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)}), 400

    # Make the protobuf from settings in order to make the URL
    else:
        try:
            # channel_set.lora_config.use_preset = 'modem_preset' in request.json.keys()
            channel_set.lora_config.modem_preset = bleach.clean(request.json.get('modem_preset')) if request.json.get('modem_preset') else None
            channel_set.lora_config.region = bleach.clean(request.json.get('lora_region')) if request.json.get('lora_region') else None
            channel_set.lora_config.hop_limit = request.json.get('lora_hop_limit') if request.json.get('lora_hop_limit') else None
            channel_set.lora_config.tx_enabled = request.json.get('lora_tx_enabled')
            channel_set.lora_config.tx_power = request.json.get('lora_tx_power') if request.json.get('lora_tx_power') else None
            channel_set.lora_config.sx126x_rx_boosted_gain = request.json.get('lora_sx126x_rx_boosted_gain')
            channel_set.lora_config.use_preset = True

            if 'psk' in request.json.keys() and request.json.get('psk'):
                channel_settings.psk = base64.urlsafe_b64decode(bleach.clean(request.json.get('psk')))
            else:
                # Zero bytes indicates no encryption
                channel_settings.psk = bytes(0)
            channel_settings.name = bleach.clean(request.json.get('name')) if request.json.get('name') else None
            channel_settings.uplink_enabled = request.json.get('uplink_enabled')
            channel_settings.downlink_enabled = request.json.get('downlink_enabled')
            channel_settings.module_settings.position_precision = request.json.get('position_precision') if request.json.get('position_precision') else None

            channel_set.settings.append(channel_settings)

        except BaseException as e:
            logger.error("Failed to save Meshtastic channel: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)}), 400

    return save_channel(channel_set)


def save_channel(channel_set: apponly_pb2.ChannelSet):
    for channel_settings in channel_set.settings:
        # Make a unique URL for each channel in a channel set
        single_channel_set = apponly_pb2.ChannelSet()
        single_channel_set.lora_config.CopyFrom(channel_set.lora_config)
        single_channel_set.settings.append(channel_settings)
        url = "https://meshtastic.org/e/#" + base64.urlsafe_b64encode(single_channel_set.SerializeToString()).decode('utf-8')

        if channel_settings.downlink_enabled and channel_settings.name not in app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS").append(channel_settings.name)
            logger.debug("Added {} to channels".format(channel_settings.name))

        meshtastic_channel_settings = MeshtasticChannel()
        meshtastic_channel_settings.psk = base64.urlsafe_b64encode(channel_settings.psk).decode('ascii')
        meshtastic_channel_settings.name = channel_settings.name
        meshtastic_channel_settings.uplink_enabled = channel_settings.uplink_enabled
        meshtastic_channel_settings.downlink_enabled = channel_settings.downlink_enabled
        meshtastic_channel_settings.position_precision = channel_settings.module_settings.position_precision
        meshtastic_channel_settings.lora_region = channel_set.lora_config.region
        meshtastic_channel_settings.lora_hop_limit = channel_set.lora_config.hop_limit
        meshtastic_channel_settings.lora_tx_enabled = channel_set.lora_config.tx_enabled
        meshtastic_channel_settings.lora_tx_power = channel_set.lora_config.tx_power
        meshtastic_channel_settings.lora_sx126x_rx_boosted_gain = channel_set.lora_config.sx126x_rx_boosted_gain
        meshtastic_channel_settings.modem_preset = channel_set.lora_config.modem_preset
        meshtastic_channel_settings.url = url

        try:
            db.session.add(meshtastic_channel_settings)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            logger.error("Channel with URL {} already exists".format(url))
            #return jsonify({'success': False, 'error': "Channel already exists"}), 400
            continue
        except BaseException as e:
            logger.error("Failed to save Meshtastic channel: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)}), 400

    return jsonify({'success': True, 'url': url})


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=['GET'])
@auth_required()
def get_channel():
    query = db.session.query(MeshtasticChannel)
    query = search(query, MeshtasticChannel, 'id')
    query = search(query, MeshtasticChannel, 'name')
    query = search(query, MeshtasticChannel, 'url')

    return paginate(query)


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=['DELETE'])
@auth_required()
def delete_channel():
    if 'url' not in request.args.keys():
        return jsonify({'success': False, 'error': gettext(u"Please provide the URL of the channel to delete")}), 400

    url = bleach.clean(request.args.get('url'))

    try:
        channel = db.session.query(MeshtasticChannel).filter(MeshtasticChannel.url == url).first()
        db.session.delete(channel)
        db.session.commit()

        if channel.name in app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            logger.error(app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"))
            app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS").pop(app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS").index(channel.name))
            logger.warning(app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"))

    except BaseException as e:
        logger.error("Failed to delete Meshtastic channel: {}".format(e))
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 400

    return jsonify({'success': True})


@meshtastic_api_blueprint.route("/api/meshtastic/generate_psk")
@auth_required()
def generate_psk():
    return jsonify({"success": True, "psk": base64.urlsafe_b64encode(os.urandom(32)).decode('ascii')}), 200



import datetime
import json
import traceback
from xml.etree.ElementTree import tostring

import pika
from bs4 import BeautifulSoup
from flask import Blueprint
import adsbxcot
import aiscot
from flask import current_app as app
from sqlalchemy import delete

from opentakserver.extensions import apscheduler, logger, db
import requests

from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.Certificate import Certificate
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.CoT import CoT
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.EUD import EUD
from opentakserver.models.GeoChat import GeoChat
from opentakserver.models.Marker import Marker
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionChange import MissionChange
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionRole import MissionRole
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.Point import Point
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.ZMIST import ZMIST
from opentakserver.functions import iso8601_string_from_datetime, datetime_from_iso8601_string, generate_delete_cot, publish_cot

scheduler_blueprint = Blueprint('scheduler_blueprint', __name__)


def get_airplanes_live_data():
    with apscheduler.app.app_context():
        try:
            r = requests.get('https://api.airplanes.live/v2/point/{}/{}/{}'
                             .format(app.config["OTS_AIRPLANES_LIVE_LAT"],
                                     app.config["OTS_AIRPLANES_LIVE_LON"],
                                     app.config["OTS_AIRPLANES_LIVE_RADIUS"]))
            if r.status_code == 200:
                rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
                rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
                rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
                channel = rabbit_connection.channel()
                channel.exchange_declare(exchange=app.config.get("OTS_ADSB_GROUP"), exchange_type="fanout")

                for craft in r.json()['ac']:
                    try:
                        event = adsbxcot.adsbx_to_cot(craft, known_craft=None)
                    except ValueError:
                        continue

                    # noinspection PyTypeChecker
                    channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
                        {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}),
                                          properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
                    # noinspection PyTypeChecker
                    channel.basic_publish(exchange='groups', routing_key=f"{app.config.get('OTS_ADSB_GROUP')}.OUT", body=json.dumps(
                        {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}),
                                          properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))

                channel.close()
                rabbit_connection.close()
            else:
                logger.error('Failed to get airplanes.live: {}'.format(r.text))
        except BaseException as e:
            logger.error("Failed to get airplanes.live: {}".format(e))
            logger.error(traceback.format_exc())


def delete_video_recordings():
    with apscheduler.app.app_context():
        VideoRecording.query.delete()
        db.session.commit()

        try:
            r = requests.get('{}/v3/recordings/list'.format(app.config.get("OTS_MEDIAMTX_API_ADDRESS")))
            if r.status_code == 200:
                recordings = r.json()
                for path in recordings['items']:
                    for recording in path['segments']:
                        r = requests.delete('{}/v3/recordings/deletesegment'.format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"),),
                                            params={'start': recording['start'], 'path': path['name']})
                        if r.status_code != 200:
                            logger.error(
                                "Failed to delete {} from {}: {}".format(recording['start'], path['name'], r.text))
        except BaseException as e:
            logger.error("Failed to delete recordings: {}".format(e))


def purge_data():
    # These are in a specific order to properly handle foreign key relationships
    delete_video_recordings()
    ZMIST.query.delete()
    VideoStream.query.delete()
    Alert.query.delete()
    CasEvac.query.delete()
    Certificate.query.delete()
    ChatroomsUids.query.delete()
    DataPackage.query.delete()
    Marker.query.delete()
    GeoChat.query.delete()
    Point.query.delete()
    RBLine.query.delete()
    Chatroom.query.delete()
    CoT.query.delete()
    MissionRole.query.delete()
    MissionContentMission.query.delete()
    MissionUID.query.delete()
    MissionChange.query.delete()
    Mission.query.delete()
    EUD.query.delete()
    Team.query.delete()
    db.session.commit()
    logger.info("Purged all data")


def get_aishub_data():
    if not app.config.get("OTS_AISHUB_USERNAME"):
        logger.error("Please set your AISHub username")
        return

    with apscheduler.app.app_context():
        try:
            params = {"username": app.config.get("OTS_AISHUB_USERNAME"), "format": 1, "output": "json"}
            if app.config.get("OTS_AISHUB_SOUTH_LAT"):
                params['latmin'] = app.config.get("OTS_AISHUB_SOUTH_LAT")
            if app.config.get("OTS_AISHUB_WEST_LON"):
                params['lonmin'] = app.config.get("OTS_AISHUB_WEST_LON")
            if app.config.get("OTS_AISHUB_NORTH_LAT"):
                params['latmax'] = app.config.get("OTS_AISHUB_NORTH_LAT")
            if app.config.get("OTS_AISHUB_EAST_LON"):
                params['lonmax'] = app.config.get("OTS_AISHUB_EAST_LON")
            if app.config.get("OTS_AISHUB_MMSI_LIST"):
                params['mmsi'] = app.config.get("OTS_AISHUB_MMSI_LIST")
            if app.config.get("OTS_AISHUB_IMO_LIST"):
                params['imo'] = app.config.get("OTS_AISHUB_IMO_LIST")
            r = requests.get("https://data.aishub.net/ws.php", params=params)

            if r.status_code != 200:
                logger.error(f"Failed to get AIS data: {r.text}")
                return

            rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
            channel = rabbit_connection.channel()
            channel.exchange_declare(exchange=app.config.get("OTS_AIS_GROUP"), exchange_type="fanout")

            for vessel in r.json()[1]:
                event = aiscot.ais_to_cot(vessel, None, None)
                # noinspection PyTypeChecker
                channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
                    {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}),
                                      properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
                # noinspection PyTypeChecker
                channel.basic_publish(exchange='groups', routing_key=f"{app.config.get('OTS_ADSB_GROUP')}.OUT", body=json.dumps(
                    {'cot': str(BeautifulSoup(event, 'xml')), 'uid': app.config['OTS_NODE_ID']}),
                                      properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))

            channel.close()
            rabbit_connection.close()
        except BaseException as e:
            logger.error(f"Failed to get AIS data: {e}")
            logger.debug(traceback.format_exc())


def delete_old_data():
    timestamp = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
        seconds=app.config.get("OTS_DELETE_OLD_DATA_SECONDS"),
        minutes=app.config.get("OTS_DELETE_OLD_DATA_MINUTES"),
        hours=app.config.get("OTS_DELETE_OLD_DATA_HOURS"),
        days=app.config.get("OTS_DELETE_OLD_DATA_DAYS"),
        weeks=app.config.get("OTS_DELETE_OLD_DATA_WEEKS"))

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()

    # I wish I hadn't made the marker's timestamp field a string...
    markers = db.session.execute(db.session.query(Marker)).all()
    for marker in markers:
        marker = marker[0]
        if datetime_from_iso8601_string(marker.production_time) <= timestamp:
            cot = generate_delete_cot(marker.uid, marker.cot.type)
            channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
                {'cot': tostring(cot).decode('utf-8'), 'uid': app.config['OTS_NODE_ID']}),
                                  properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
            db.session.delete(marker)

    alerts = db.session.execute(db.session.query(Alert).where(Alert.start_time <= timestamp)).all()
    for alert in alerts:
        alert = alert[0]
        publish_cot(generate_delete_cot(alert.uid, alert.cot.type), channel)
        db.session.delete(alert)

    rb_lines = db.session.execute(db.session.query(RBLine).where(RBLine.timestamp <= timestamp)).all()
    for rb_line in rb_lines:
        rb_line = rb_line[0]
        publish_cot(generate_delete_cot(rb_line.uid, rb_line.cot.type), channel)
        db.session.delete(rb_line)

    db.session.execute(delete(GeoChat).where(GeoChat.timestamp <= timestamp))

    timestamp = datetime.datetime.now() - datetime.timedelta(
        seconds=app.config.get("OTS_DELETE_OLD_DATA_SECONDS"),
        minutes=app.config.get("OTS_DELETE_OLD_DATA_MINUTES"),
        hours=app.config.get("OTS_DELETE_OLD_DATA_HOURS"),
        days=app.config.get("OTS_DELETE_OLD_DATA_DAYS"),
        weeks=app.config.get("OTS_DELETE_OLD_DATA_WEEKS"))

    db.session.execute(delete(EUD).where(EUD.last_event_time <= timestamp))
    db.session.execute(delete(Point).where(Point.timestamp <= timestamp))
    db.session.execute(delete(CoT).where(CoT.timestamp <= timestamp))
    db.session.commit()

    channel.close()
    rabbit_connection.close()

    logger.info(f"Deleted data older than {iso8601_string_from_datetime(timestamp)}")



from gevent import monkey
monkey.patch_all()

import pytz
from opentakserver.models.role import Role
from opentakserver.models.Group import Group, GroupTypeEnum

from opentakserver.UsernameValidator import UsernameValidator

import sys
import traceback
import logging

from flask_migrate import Migrate, upgrade
from opentakserver.PasswordValidator import PasswordValidator

import platform
import requests
from sqlalchemy import insert
import sqlite3
from opentakserver.models.Icon import Icon
from opentakserver.plugins.Plugin import Plugin
from opentakserver.plugins.PluginManager import PluginManager
from opentakserver.sql_jobstore import SQLJobStore

import yaml

from opentakserver.EmailValidator import EmailValidator

from logging.handlers import TimedRotatingFileHandler
import os

import colorlog
from werkzeug.middleware.proxy_fix import ProxyFix
from datetime import datetime, timezone
import sqlalchemy

import flask_wtf

import pika
from flask import Flask, current_app, g, request, session
from flask_cors import CORS

from flask_security import Security, SQLAlchemyUserDatastore, hash_password, uia_username_mapper, uia_email_mapper
from flask_security.models import fsqla_v3 as fsqla, fsqla_v3
from flask_security.signals import user_registered

import opentakserver
from opentakserver.extensions import logger, db, socketio, mail, apscheduler, ldap_manager, babel
from opentakserver.defaultconfig import DefaultConfig
from opentakserver.models.WebAuthn import WebAuthn

from opentakserver.controllers.meshtastic_controller import MeshtasticController
from opentakserver.certificate_authority import CertificateAuthority

try:
    from opentakserver.mumble.mumble_ice_app import MumbleIceDaemon
except ModuleNotFoundError:
    print("Mumble auth not supported on this platform")


def get_locale():
    if 'language' in session:
        return session['language']
    return request.accept_languages.best_match(current_app.config.get("OTS_LANGUAGES").keys())


def get_timezone():
    # Always return UTC and let the frontend handle converting timezones
    return pytz.timezone("UTC")


def init_extensions(app):
    db.init_app(app)
    Migrate(app, db)

    logger.info(f"OpenTAKServer {opentakserver.__version__}")
    logger.info("Loading the database...")
    with app.app_context():
        upgrade(directory=os.path.join(os.path.dirname(os.path.realpath(opentakserver.__file__)), 'migrations'))
        # Flask-Migrate does weird things to the logger
        logger.disabled = False
        logger.parent.handlers.pop()
        if app.config.get("DEBUG"):
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    # Handle config options that can't be serialized to yaml
    app.config.update({"SCHEDULER_JOBSTORES": {'default': SQLJobStore(url=app.config.get("SQLALCHEMY_DATABASE_URI"))}})
    identity_attributes = [{"username": {"mapper": uia_username_mapper, "case_insensitive": True}}]

    # Don't allow registration unless email is enabled
    if app.config.get("OTS_ENABLE_EMAIL"):
        identity_attributes.append({"email": {"mapper": uia_email_mapper, "case_insensitive": True}})
        app.config.update({
            "SECURITY_REGISTERABLE": True,
            "SECURITY_CONFIRMABLE": True,
            "SECURITY_RECOVERABLE": True,
            "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator", "email"]
        })
    else:
        app.config.update({
            "SECURITY_REGISTERABLE": False,
            "SECURITY_CONFIRMABLE": False,
            "SECURITY_RECOVERABLE": False,
            "SECURITY_TWO_FACTOR_ENABLED_METHODS": ["authenticator"]
        })

    if app.config.get("OTS_ENABLE_LDAP"):
        logger.info("Enabling LDAP")
        ldap_manager.init_app(app)
        identity_attributes.append({"ldap": {}})

    app.config.update({"SECURITY_USER_IDENTITY_ATTRIBUTES": identity_attributes})

    ca = CertificateAuthority(logger, app)
    ca.create_ca()

    cors = CORS(app, resources={r"/api/*": {"origins": "*"}, r"/Marti/*": {"origins": "*"}, r"/*": {"origins": "*"}},
                supports_credentials=True)
    flask_wtf.CSRFProtect(app)

    socketio_logger = False
    if app.config.get("DEBUG"):
        socketio_logger = logger
    socketio.init_app(app, logger=socketio_logger, ping_timeout=1, message_queue="amqp://" + app.config.get("OTS_RABBITMQ_SERVER_ADDRESS"))

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))

    channel = rabbit_connection.channel()
    channel.exchange_declare('dms', durable=True, exchange_type='direct')
    channel.exchange_declare('cot_parser', durable=True, exchange_type='direct')
    channel.exchange_declare('chatrooms', durable=True, exchange_type='direct')
    channel.exchange_declare("missions", durable=True, exchange_type='topic')  # For Data Sync mission feeds
    channel.exchange_declare("groups", durable=True, exchange_type='topic')  # For channels/groups
    channel.close()
    rabbit_connection.close()

    if not apscheduler.running:
        apscheduler.init_app(app)
        apscheduler.start(paused=False)

    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.user import User
    from opentakserver.models.role import Role

    user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
    app.security = Security(app, user_datastore, mail_util_cls=EmailValidator, password_util_cls=PasswordValidator, username_util_cls=UsernameValidator)

    mail.init_app(app)

    babel.init_app(app, locale_selector=get_locale, timezone_selector=get_timezone)


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG
    logger.setLevel(level)

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            '%(log_color)s[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s', datefmt="%Y-%m-%d %H:%M:%S")
        color_log_handler.setFormatter(color_log_formatter)
        logger.addHandler(color_log_handler)
        logger.info("Added color logger")

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'logs', 'opentakserver.log'),
                                  when=app.config.get("OTS_LOG_ROTATE_WHEN"), interval=app.config.get("OTS_LOG_ROTATE_INTERVAL"),
                                  backupCount=app.config.get("OTS_BACKUP_COUNT"))
    fh.setFormatter(logging.Formatter("[%(asctime)s] - OpenTAKServer[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s"))
    logger.addHandler(fh)


def create_app(cli=True):
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)
    setup_logging(app)

    if not cli:
        # Load config.yml if it exists
        if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
            app.config.from_file(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load)
        else:
            # First run, created config.yml based on default settings
            logger.info("Creating config.yml")
            with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
                conf = {}
                for option in DefaultConfig.__dict__:
                    # Fix the sqlite DB path on Windows
                    if option == "SQLALCHEMY_DATABASE_URI" and platform.system() == "Windows" and DefaultConfig.__dict__[option].startswith("sqlite"):
                        conf[option] = DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                    elif option.isupper():
                        conf[option] = DefaultConfig.__dict__[option]
                config.write(yaml.safe_dump(conf))

        # Try to set the MediaMTX token
        if app.config.get("OTS_MEDIAMTX_ENABLE"):
            try:
                new_conf = None
                with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"), "r") as mediamtx_config:
                    conf = mediamtx_config.read()
                    if "MTX_TOKEN" in conf:
                        new_conf = conf.replace("MTX_TOKEN", app.config.get("OTS_MEDIAMTX_TOKEN"))
                if new_conf:
                    with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "mediamtx.yml"), "w") as mediamtx_config:
                        mediamtx_config.write(new_conf)
            except BaseException as e:
                logger.error("Failed to set MediaMTX token: {}".format(e))
        else:
            logger.info("MediaMTX disabled")

        init_extensions(app)

        from opentakserver.blueprints.marti_api import marti_blueprint
        app.register_blueprint(marti_blueprint)

        from opentakserver.blueprints.ots_api import ots_api
        app.register_blueprint(ots_api)

        from opentakserver.blueprints.ots_socketio import ots_socketio_blueprint
        app.register_blueprint(ots_socketio_blueprint)

        from opentakserver.blueprints.scheduled_jobs import scheduler_blueprint
        app.register_blueprint(scheduler_blueprint)

        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_host=1)

    else:
        from opentakserver.blueprints.cli import ots, translate
        app.cli.add_command(ots, name="ots")
        app.cli.add_command(translate, name="translate")

        if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
            app.config.from_file(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load)
            db.init_app(app)
            Migrate(app, db)

        flask_wtf.CSRFProtect(app)

        try:
            fsqla.FsModels.set_db_info(db)
        except sqlalchemy.exc.InvalidRequestError:
            pass

        from opentakserver.models.user import User
        from opentakserver.models.role import Role

        user_datastore = SQLAlchemyUserDatastore(db, User, Role, WebAuthn)
        app.security = Security(app, user_datastore, mail_util_cls=EmailValidator, password_util_cls=PasswordValidator, username_util_cls=UsernameValidator)

        # Register blueprints to properly import all the DB models without circular imports
        from opentakserver.blueprints.marti_api import marti_blueprint
        app.register_blueprint(marti_blueprint)

        from opentakserver.blueprints.ots_api import ots_api
        app.register_blueprint(ots_api)

        from opentakserver.blueprints.ots_socketio import ots_socketio_blueprint
        app.register_blueprint(ots_socketio_blueprint)

        from opentakserver.blueprints.scheduled_jobs import scheduler_blueprint
        app.register_blueprint(scheduler_blueprint)

    return app


@user_registered.connect_via(current_app)
def user_registered_sighandler(app, user, confirmation_token, **kwargs):
    default_role = app.security.datastore.find_or_create_role(
        name="user", permissions={"user-read", "user-write"}
    )
    app.security.datastore.add_role_to_user(user, default_role)


def create_default_groups(app):
    with app.app_context():
        if not app.config.get("OTS_ENABLE_LDAP"):
            anon_group = db.session.execute(db.session.query(Group).filter_by(name="__ANON__")).first()
            adsb_group = db.session.execute(db.session.query(Group).filter_by(name=app.config.get("OTS_ADSB_GROUP"))).first()
            ais_group = db.session.execute(db.session.query(Group).filter_by(name=app.config.get("OTS_AIS_GROUP"))).first()
            meshtastic_group = db.session.execute(db.session.query(Group).filter_by(name=app.config.get("OTS_MESHTASTIC_GROUP"))).first()

            # Commit to DB after every one to ensure that get_next_bitpos works

            if not anon_group:
                logger.info("Creating the __ANON__ group")
                anon_group = Group()
                anon_group.name = "__ANON__"
                anon_group.type = GroupTypeEnum.SYSTEM
                anon_group.bitpos = 2
                db.session.add(anon_group)
                db.session.commit()

            if not adsb_group:
                logger.info(f"Creating the {app.config.get('OTS_ADSB_GROUP')} group")
                adsb_group = Group()
                adsb_group.name = app.config.get("OTS_ADSB_GROUP")
                adsb_group.type = GroupTypeEnum.SYSTEM
                adsb_group.bitpos = adsb_group.get_next_bitpos()
                db.session.add(adsb_group)
                db.session.commit()

            if not ais_group:
                logger.info(f"Creating the {app.config.get('OTS_AIS_GROUP')} group")
                ais_group = Group()
                ais_group.name = app.config.get("OTS_AIS_GROUP")
                ais_group.type = GroupTypeEnum.SYSTEM
                ais_group.bitpos = ais_group.get_next_bitpos()
                db.session.add(ais_group)
                db.session.commit()

            if not meshtastic_group:
                logger.info(f"Creating the {app.config.get('OTS_MESHTASTIC_GROUP')} group")
                meshtastic_group = Group()
                meshtastic_group.name = app.config.get("OTS_MESHTASTIC_GROUP")
                meshtastic_group.type = GroupTypeEnum.SYSTEM
                meshtastic_group.bitpos = meshtastic_group.get_next_bitpos()
                db.session.add(meshtastic_group)
                db.session.commit()


def main(app):
    with app.app_context():
        # Download the icon sets if they aren't already in the DB
        icons = db.session.query(Icon).count()
        if icons == 0:
            logger.info("Downloading icons...")
            try:
                r = requests.get("https://github.com/brian7704/OpenTAKServer-Installer/raw/master/iconsets.sqlite", stream=True)
                with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite"), "wb") as f:
                    f.write(r.content)

                def dict_factory(cursor, row):
                    d = {}
                    for idx, col in enumerate(cursor.description):
                        d[col[0]] = row[idx]
                    return d

                con = sqlite3.connect(os.path.join(app.config.get("OTS_DATA_FOLDER"), "icons.sqlite"))
                con.row_factory = dict_factory
                cur = con.cursor()
                rows = cur.execute("SELECT * FROM icons")
                for row in rows:
                    db.session.execute(insert(Icon).values(**row))
                db.session.commit()
            except BaseException as e:
                logger.error("Failed to download icons: {}".format(e))
                logger.debug(traceback.format_exc())

        if app.config.get("DEBUG"):
            logger.debug("Starting in debug mode")
        else:
            logger.info("Starting in production mode")

        app.security.datastore.find_or_create_role(
            name="user", permissions={"user-read", "user-write"}
        )

        app.security.datastore.find_or_create_role(
            name="administrator", permissions={"administrator"}
        )

        # Make sure at least one admin user exists
        admin_user = db.session.execute(db.session.query(Role).join(fsqla_v3.FsModels.roles_users).where(Role.name == "administrator")).scalar()
        if not admin_user:
            logger.info("Creating administrator account. The password is 'password'")
            app.security.datastore.create_user(username="administrator",
                                               password=hash_password("password"), roles=["administrator"])
        db.session.commit()

    if app.config.get("OTS_ENABLE_MESHTASTIC"):
        mestastic_thread = MeshtasticController(app.app_context())
        app.mestastic_thread = mestastic_thread
    else:
        app.meshtastic_thread = None

    if app.config.get("OTS_ENABLE_MUMBLE_AUTHENTICATION"):
        try:
            logger.info("Starting Mumble authentication handler")
            mumble_daemon = MumbleIceDaemon(app, logger)
            mumble_daemon.daemon = True
            mumble_daemon.start()
        except BaseException as e:
            logger.error("Failed to enable Mumble authentication: {}".format(e))
            logger.error(traceback.format_exc())
    else:
        logger.info("Mumble authentication handler disabled")

    if app.config.get("OTS_ENABLE_PLUGINS"):
        try:
            app.plugin_manager = PluginManager(Plugin.group, app)
            app.plugin_manager.load_plugins()
            app.plugin_manager.activate(app)
        except BaseException as e:
            logger.error(f"Failed to load plugins: {e}")
            logger.debug(traceback.format_exc())

    app.start_time = datetime.now(timezone.utc)

    create_default_groups(app)

    try:
        socketio.run(app, host=app.config.get("OTS_LISTENER_ADDRESS"), port=app.config.get("OTS_LISTENER_PORT"),
                     debug=app.config.get("DEBUG"), log_output=app.config.get("DEBUG"), use_reloader=False)
    except KeyboardInterrupt:
        logger.warning("Caught CTRL+C, exiting...")
        if app.config.get("OTS_ENABLE_PLUGINS"):
            app.plugin_manager.stop_plugins()


def start():
    app = create_app(cli=False)
    main(app)



import io
import os
import random
import re
import shutil
import subprocess
import traceback
import uuid
import zipfile
from pathlib import Path
from shutil import copyfile, rmtree
from urllib.parse import urlparse

from flask import request
from jinja2 import Template
from .ca_config import ca_config, server_config


class CertificateAuthority:

    def __init__(self, logger, app):
        self.logger = logger
        self.app = app

    def create_ca(self):
        if not self.check_if_ca_exists():
            self.logger.info("Creating CA...")
            os.makedirs(self.app.config.get("OTS_CA_FOLDER"), exist_ok=True)

            f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca_config.cfg"), 'w')
            f.write(ca_config)
            f.close()

            subject = self.app.config.get("OTS_CA_SUBJECT") + "/CN={}".format(self.app.config.get("OTS_CA_NAME"))

            command = (
                'openssl req -new -sha256 -x509 -days {} -extensions v3_ca -keyout {} -out {} -passout pass:{} -config {} -subj {}'
                .format(self.app.config.get("OTS_CA_EXPIRATION_TIME"),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca-do-not-share.key"),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                        self.app.config.get("OTS_CA_PASSWORD"),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca_config.cfg"),
                        subject))

            self.logger.debug(command)

            exit_code = subprocess.call(command, shell=True)

            if exit_code:
                raise Exception("Failed to create ca.pem. Exit code {}".format(exit_code))

            command = ('openssl x509 -in {} -addtrust clientAuth -addtrust serverAuth -setalias {} -out {}'
                       .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                               self.app.config.get("OTS_CA_NAME"),
                               os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca-trusted.pem")))

            self.logger.debug(command)

            exit_code = subprocess.call(command, shell=True)

            if exit_code:
                raise Exception("Failed to add trust to CA. Exit code {}".format(exit_code))

            use_legacy = not subprocess.call("openssl list -providers", shell=True)

            if use_legacy:
                command = ('openssl pkcs12 -legacy -export -in {} -out {} -passout pass:{} -nokeys -caname {}'
                           .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca-trusted.pem"),
                                   os.path.join(self.app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"),
                                   self.app.config.get("OTS_CA_PASSWORD"),
                                   self.app.config.get("OTS_CA_NAME")))
            else:
                command = ('openssl pkcs12 -export -in {} -out {} -passout pass:{} -nokeys -caname {}'
                           .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca-trusted.pem"),
                                   os.path.join(self.app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"),
                                   self.app.config.get("OTS_CA_PASSWORD"),
                                   self.app.config.get("OTS_CA_NAME")))

            self.logger.debug(command)

            exit_code = subprocess.call(command, shell=True)

            if exit_code:
                raise Exception("Failed to export truststore. Exit code {}".format(exit_code))

            Path(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "crl_index.txt")).touch()
            f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "crl_index.txt.attr"), 'w')
            f.write("unique_subject = no")
            f.close()

            command = ('cd {} && openssl ca -config {} -gencrl -keyfile {} -passin pass:{} -cert {} -out {}'
                       .format(self.app.config.get("OTS_CA_FOLDER"),
                               os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca_config.cfg"),
                               os.path.join(self.app.config.get("OTS_CA_FOLDER"), 'ca-do-not-share.key'),
                               self.app.config.get("OTS_CA_PASSWORD"),
                               os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                               os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.crl")))

            self.logger.debug(command)

            exit_code = subprocess.call(command, shell=True)

            if exit_code:
                raise Exception("Failed to create crl. Exit code {}".format(exit_code))

            self.logger.debug("Creating server cert...")
            self.issue_certificate("opentakserver", True)
            self.logger.info(
                "Certificate authority created successfully. You may need to restart nginx if it's proxying SSL requests.")

        else:
            self.logger.debug("CA already exists")

    def issue_certificate(self, common_name, server=False):
        if not os.path.exists(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem")):
            raise FileNotFoundError("ca.pem not found")

        if os.path.exists(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name)):
            shutil.rmtree(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name))

        os.makedirs(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name))

        subject = self.app.config.get("OTS_CA_SUBJECT") + "/CN={}".format(common_name)

        command = (
            'openssl req -new -newkey rsa:2048 -sha256 -keyout {} -passout pass:{} -out {} -subj {} -config {}'
            .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".key"),
                    self.app.config.get("OTS_CA_PASSWORD"),
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"),
                    subject,
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca_config.cfg")))

        self.logger.debug(command)

        exit_code = subprocess.call(command, shell=True)
        if exit_code:
            raise Exception("Failed to create csr. Exit code {}".format(exit_code))

        csr = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"), 'r')
        csr_bytes = csr.read().encode()
        csr.close()

        self.sign_csr(csr_bytes, common_name, server)

        use_legacy = not subprocess.call("openssl list -providers", shell=True)

        if use_legacy:
            command = (
                'openssl pkcs12 -legacy -export -in {}.pem -inkey {}.key -out {}.p12 -name {} -CAfile {} -passin pass:{} -passout pass:{}'
                .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        common_name,
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                        self.app.config.get("OTS_CA_PASSWORD"),
                        self.app.config.get("OTS_CA_PASSWORD")))
        else:
            command = (
                'openssl pkcs12 -export -in {}.pem -inkey {}.key -out {}.p12 -name {} -CAfile {} -passin pass:{} -passout pass:{}'
                .format(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name),
                        common_name,
                        os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                        self.app.config.get("OTS_CA_PASSWORD"),
                        self.app.config.get("OTS_CA_PASSWORD")))

        self.logger.debug(command)

        exit_code = subprocess.call(command, shell=True)
        if exit_code:
            raise Exception("Failed to export p12 key. Exit code {}".format(exit_code))

        os.chmod(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".key"), 0o620)

        command = 'openssl rsa -in {} -passin pass:{} -out {}'.format(
            os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".key"),
            os.path.join(self.app.config.get("OTS_CA_PASSWORD")),
            os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + '.nopass.key'))

        self.logger.debug(command)

        exit_code = subprocess.call(command, shell=True)
        if exit_code:
            raise Exception("Failed to remove server key password. Exit code {}".format(exit_code))

        if not server:
            return self.generate_zip(common_name)
        else:
            # Generate public key for PyJWT to validate tokens
            command = "openssl x509 -pubkey -in {} -out {}".format(
                os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"),
                os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pub"))

            self.logger.debug(command)

            exit_code = subprocess.call(command, shell=True)
            if exit_code:
                raise Exception("Failed to generate server's public key. Exit code {}".format(exit_code))

    def sign_csr(self, csr_bytes, common_name, server=False):
        os.makedirs(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name), exist_ok=True)
        f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"), 'wb')
        f.write(csr_bytes)
        f.close()

        if server:
            if re.match("^[0-9]{1,3}.[0-9]{1,3}.[0-9]{1,3}.[0-9]{1,3}$", common_name):
                alt_name_field = "IP.1"
            else:
                alt_name_field = "DNS.1"

            f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                  "{}_config.cfg".format(common_name)), 'w')

            f.write(server_config.render(alt_name_field=alt_name_field, common_name=common_name))

            f.close()

            config_file = os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                       "{}_config.cfg".format(common_name))
            extensions = 'server'
        else:
            config_file = os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca_config.cfg")
            extensions = 'client'

        command = (
            'openssl x509 -sha256 -req -days {} -in {} -CA {} -CAkey {} -out {} -set_serial {} -passin pass:{} -extensions {} -extfile {}'
            .format(self.app.config.get("OTS_CA_EXPIRATION_TIME"),
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr"),
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"),
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca-do-not-share.key"),
                    os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"),
                    random.randint(10000, 100000),
                    self.app.config.get("OTS_CA_PASSWORD"),
                    extensions,
                    config_file))

        self.logger.debug(command)

        exit_code = subprocess.call(command, shell=True)
        if exit_code:
            raise Exception("Failed to sign csr. Exit code {}".format(exit_code))

        if server:
            with open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "ca.pem"), 'r') as ca_file:
                f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"), 'a')
                f.write(ca_file.read())
                f.close()

        f = open(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".pem"), 'r')
        cert_bytes = f.read().encode()
        f.close()

        return cert_bytes

    def check_if_ca_exists(self):
        return os.path.exists(os.path.join(self.app.config.get("OTS_CA_FOLDER"), 'ca.pem'))

    def generate_zip(self, common_name):
        truststore = os.path.join(self.app.config.get("OTS_CA_FOLDER"), 'truststore-root.p12')
        user_p12 = os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                "{}.p12".format(common_name))
        user_file_path = os.path.join(self.app.config.get("OTS_CA_FOLDER"), "certs", common_name)
        random_id = uuid.uuid4()
        new_uid = uuid.uuid4()
        parent_folder = "80b828699e074a239066d454a76284eb"
        folder = "5c2bfcae3d98c9f4d262172df99ebac5"

        pref_file_template = Template("""<?xml version='1.0' standalone='yes'?>
                <preferences>
                    <preference version="1" name="cot_streams">
                        <entry key="count" class="class java.lang.Integer">1</entry>
                        <entry key="description0" class="class java.lang.String">OpenTAKServer_{{ server }}</entry>
                        <entry key="enabled0" class="class java.lang.Boolean">true</entry>
                        <entry key="connectString0" class="class java.lang.String">{{ server }}:{{ ssl_port }}:ssl</entry>
                    </preference>
                    <preference version="1" name="com.atakmap.app_preferences">
                        <entry key="deviceProfileEnableOnConnect" class="class java.lang.Boolean">true</entry>
                        <entry key="displayServerConnectionWidget" class="class java.lang.Boolean">true</entry>
                        <entry key="caLocation" class="class java.lang.String">/storage/emulated/0/atak/cert/{{ server_filename }}</entry>
                        <entry key="caPassword" class="class java.lang.String">{{ cert_password }}</entry>
                        <entry key="clientPassword" class="class java.lang.String">{{ cert_password }}</entry>
                        <entry key="certificateLocation" class="class java.lang.String">/storage/emulated/0/atak/cert/{{ user_filename }}</entry>
                        <entry key="appMgmtEnableUpdateServer" class="class java.lang.Boolean">true</entry>
                        <entry key="atakUpdateServerUrl" class="class java.lang.String">https://{{ server }}:{{ marti_port }}/api/packages</entry>
                        <entry key="repoStartupSync" class="class java.lang.Boolean">true</entry>
                        <entry key="updateServerCaLocation" class="class java.lang.String">/storage/emulated/0/atak/cert/{{ server_filename }}</entry>
                        <entry key="updateServerCaPassword" class="class java.lang.String">{{ cert_password }}</entry>
                    </preference>
                </preferences>
                """)

        manifest_file_template = Template("""<MissionPackageManifest version="2">
                   <Configuration>
                      <Parameter name="uid" value="{{ uid }}"/>
                      <Parameter name="name" value="OpenTAKServer_{{ server }}"/>
                      <Parameter name="onReceiveDelete" value="true"/>
                   </Configuration>
                   <Contents>
                      <Content ignore="false" zipEntry="{{ folder }}/preference.pref"/>
                      <Content ignore="false" zipEntry="{{ folder }}/{{ server_filename }}"/>
                      <Content ignore="false" zipEntry="{{ folder }}/{{ user_filename }}"/>	  
                   </Contents>
                </MissionPackageManifest>
                """)

        manifest_file_parent_template = Template("""<MissionPackageManifest version="2">
                       <Configuration>
                          <Parameter name="uid" value="{{ uid }}"/>
                          <Parameter name="name" value="OpenTAKServer_{{ server }}_CONFIG"/>
                       </Configuration>
                       <Contents>
                          <Content ignore="false" zipEntry="{{ folder }}/{{ internal_dp_name }}.zip"/>
                       </Contents>
                    </MissionPackageManifest>
                    """)

        pref = pref_file_template.render(server=urlparse(request.url_root).hostname,
                                         marti_port=self.app.config.get('OTS_MARTI_HTTPS_PORT'),
                                         server_filename="truststore-root.p12",
                                         user_filename=f"{common_name}.p12",
                                         cert_password=self.app.config.get("OTS_CA_PASSWORD"),
                                         ssl_port=self.app.config.get("OTS_SSL_STREAMING_PORT"))
        man = manifest_file_template.render(uid=random_id, server=urlparse(request.url_root).hostname,
                                            server_filename="truststore-root.p12",
                                            user_filename=f"{common_name}.p12", folder=folder)
        man_parent = manifest_file_parent_template.render(uid=new_uid, server=urlparse(request.url_root).hostname,
                                                          folder=parent_folder,
                                                          internal_dp_name=common_name)

        if not os.path.exists(os.path.join(user_file_path, folder)):
            os.makedirs(os.path.join(user_file_path, folder))

        if not os.path.exists(os.path.join(user_file_path, 'MANIFEST')):
            os.makedirs(os.path.join(user_file_path, 'MANIFEST'))

        with open(os.path.join(user_file_path, folder, 'preference.pref'), 'w') as pref_file:
            pref_file.write(pref)

        with open(os.path.join(user_file_path, 'MANIFEST', 'manifest.xml'), 'w') as manifest_file:
            manifest_file.write(man)

        self.logger.debug("Generating inner Data Package: {}.zip".format(common_name))

        copyfile(truststore, os.path.join(user_file_path, folder, "truststore-root.p12"))
        self.logger.debug("Copying {} to {}".format(truststore, os.path.join(user_file_path, folder,
                                                                             "truststore-root.p12")))
        copyfile(user_p12, os.path.join(user_file_path, folder, "{}.p12".format(common_name)))
        zipf = zipfile.ZipFile(os.path.join(user_file_path, "{}.zip".format(common_name)), 'w', zipfile.ZIP_DEFLATED)

        os.chdir(os.path.join(user_file_path))

        for root, dirs, files in os.walk(folder):
            for file in files:
                zipf.write(os.path.join(root, file))
        for root, dirs, files in os.walk('MANIFEST'):
            for file in files:
                self.logger.debug("adding {} to zip".format(os.path.join(root, file)))
                zipf.write(os.path.join(root, file))
        zipf.close()

        rmtree(os.path.join(user_file_path, "MANIFEST"))
        rmtree(os.path.join(user_file_path, folder))

        # Create outer DP...because WinTAK
        if not os.path.exists(os.path.join(user_file_path, parent_folder)):
            os.makedirs(os.path.join(user_file_path, parent_folder))
        if not os.path.exists(os.path.join(user_file_path, "MANIFEST")):
            os.makedirs(os.path.join(user_file_path, "MANIFEST"))
        with open(os.path.join(user_file_path, "MANIFEST", 'manifest.xml'), 'w') as manifest_parent:
            manifest_parent.write(man_parent)

        self.logger.info("Generating Main Data Package: {}_CONFIG.zip".format(common_name))
        copyfile(os.path.join(user_file_path, "{}.zip".format(common_name)), os.path.join(user_file_path, parent_folder,
                                                                                          "{}.zip".format(common_name)))
        zipp = zipfile.ZipFile(os.path.join(user_file_path, "{}_CONFIG.zip".format(common_name)), 'w',
                               zipfile.ZIP_DEFLATED)

        for root, dirs, files in os.walk(parent_folder):
            for file in files:
                zipp.write(os.path.join(root, file))
        for root, dirs, files in os.walk('MANIFEST'):
            for file in files:
                zipp.write(os.path.join(root, file))
        zipp.close()

        # Generate iTAK zip
        itak_preferences = Template("""<?xml version='1.0' standalone='yes'?>
<preferences>
  <preference version="1" name="cot_streams">
    <entry key="count" class="class java.lang.Integer">1</entry>
    <entry key="description0" class="class java.lang.String">OpenTAKServer_{{ server }}</entry>
    <entry key="enabled0" class="class java.lang.Boolean">true</entry>
    <entry key="connectString0" class="class java.lang.String">{{ server }}:{{ ssl_port }}:ssl</entry>
  </preference>
  <preference version="1" name="com.atakmap.app_preferences">
    <entry key="displayServerConnectionWidget" class="class java.lang.Boolean">true</entry>
    <entry key="caLocation" class="class java.lang.String">cert/truststore-root.p12</entry>
    <entry key="caPassword" class="class java.lang.String">{{ cert_password }}</entry>
    <entry key="clientPassword" class="class java.lang.String">{{ cert_password }}</entry>
    <entry key="certificateLocation" class="class java.lang.String">cert/{{ common_name }}.p12</entry>
  </preference>
</preferences>

""")

        f = open(os.path.join(user_file_path, "config.pref"), 'w')
        f.write(itak_preferences.render(server=urlparse(request.url_root).hostname,
                                        ssl_port=self.app.config.get("OTS_SSL_STREAMING_PORT"),
                                        cert_password=self.app.config.get("OTS_CA_PASSWORD"),
                                        common_name=common_name))
        f.close()

        self.logger.info("Generating {}_CONFIG_iTAK.zip...".format(common_name))
        itak_zip = zipfile.ZipFile(os.path.join(user_file_path, "{}_CONFIG_iTAK.zip".format(common_name)), 'w',
                                   zipfile.ZIP_DEFLATED)
        itak_zip.write(os.path.join(user_file_path, "config.pref"), "config.pref")
        itak_zip.write(os.path.join(user_file_path, common_name + ".p12"), common_name + ".p12")
        itak_zip.write(os.path.join(self.app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"), "truststore-root.p12")
        itak_zip.close()

        rmtree(os.path.join(user_file_path, "MANIFEST"))
        rmtree(os.path.join(user_file_path, parent_folder))
        os.remove(os.path.join(user_file_path, "{}.zip".format(common_name)))
        os.remove(os.path.join(user_file_path, "config.pref".format(common_name)))

        return ["{}_CONFIG.zip".format(common_name), "{}_CONFIG_iTAK.zip".format(common_name)]



"""Added some cascades

Revision ID: 2b5bddd207ae
Revises: 298430e7849d
Create Date: 2024-11-07 14:34:56.132740

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2b5bddd207ae'
down_revision = '298430e7849d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table('alerts', schema=None) as batch_op:
            batch_op.drop_constraint('alert_cot', type_='foreignkey')
    except:
        pass
    try:
        with op.batch_alter_table('alerts', schema=None) as batch_op:
            batch_op.drop_constraint('alert_point', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('alerts', schema=None) as batch_op:
            batch_op.drop_constraint('alert_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('alerts', schema=None) as batch_op:
        batch_op.create_foreign_key('alert_point', 'points', ['point_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key('alert_eud', 'euds', ['sender_uid'], ['uid'], ondelete='CASCADE')
        batch_op.create_foreign_key('alert_cot', 'cot', ['cot_id'], ['id'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('casevac', schema=None) as batch_op:
            batch_op.drop_constraint('casevac_cot', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('casevac', schema=None) as batch_op:
            batch_op.drop_constraint('casevac_point', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('casevac', schema=None) as batch_op:
            batch_op.drop_constraint('casevac_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('casevac', schema=None) as batch_op:
        batch_op.create_foreign_key('casevac_cot', 'cot', ['cot_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key('casevac_eud', 'euds', ['sender_uid'], ['uid'], ondelete='CASCADE')
        batch_op.create_foreign_key('casevac_point', 'points', ['point_id'], ['id'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('certificates', schema=None) as batch_op:
            batch_op.drop_constraint('certificate_eud', type_='foreignkey')
    except BaseException as e:
        print(f"Shit failed {e}")

    with op.batch_alter_table('certificates', schema=None) as batch_op:
        batch_op.create_foreign_key('certificate_eud', 'euds', ['eud_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('markers', schema=None) as batch_op:
            batch_op.drop_constraint('marker_cot', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('markers', schema=None) as batch_op:
            batch_op.drop_constraint('marker_point', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('markers', schema=None) as batch_op:
            batch_op.drop_constraint('marker_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('markers', schema=None) as batch_op:
        batch_op.create_foreign_key('markers_cot', 'cot', ['cot_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key('markers_eud', 'euds', ['parent_uid'], ['uid'], ondelete='CASCADE')
        batch_op.create_foreign_key('markers_point', 'points', ['point_id'], ['id'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('points', schema=None) as batch_op:
            batch_op.drop_constraint('point_cot', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('points', schema=None) as batch_op:
            batch_op.drop_constraint('point_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('points', schema=None) as batch_op:
        batch_op.create_foreign_key('point_cot', 'cot', ['cot_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key('point_eud', 'euds', ['device_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('rb_lines', schema=None) as batch_op:
            batch_op.drop_constraint('rb_line_cot', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('rb_lines', schema=None) as batch_op:
            batch_op.drop_constraint('rb_line_point', type_='foreignkey')
    except:
        pass

    try:
        with op.batch_alter_table('rb_lines', schema=None) as batch_op:
            batch_op.drop_constraint('rb_line_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('rb_lines', schema=None) as batch_op:
        batch_op.create_foreign_key('rb_line_cot', 'cot', ['cot_id'], ['id'], ondelete='CASCADE')
        batch_op.create_foreign_key('rb_line_eud', 'euds', ['sender_uid'], ['uid'], ondelete='CASCADE')
        batch_op.create_foreign_key('rb_line_point', 'points', ['point_id'], ['id'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('zmist', schema=None) as batch_op:
            batch_op.drop_constraint('zmist_casevac', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('zmist', schema=None) as batch_op:
        batch_op.create_foreign_key('zmist_casevac', 'casevac', ['casevac_uid'], ['uid'], ondelete='CASCADE')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('zmist', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'casevac', ['casevac_uid'], ['uid'])

    with op.batch_alter_table('rb_lines', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'euds', ['sender_uid'], ['uid'])
        batch_op.create_foreign_key(None, 'cot', ['cot_id'], ['id'])
        batch_op.create_foreign_key(None, 'points', ['point_id'], ['id'])

    with op.batch_alter_table('points', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'cot', ['cot_id'], ['id'])
        batch_op.create_foreign_key(None, 'euds', ['device_uid'], ['uid'])

    with op.batch_alter_table('markers', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'cot', ['cot_id'], ['id'])
        batch_op.create_foreign_key(None, 'points', ['point_id'], ['id'])
        batch_op.create_foreign_key(None, 'euds', ['parent_uid'], ['uid'])

    with op.batch_alter_table('certificates', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'euds', ['eud_uid'], ['uid'])

    with op.batch_alter_table('casevac', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'cot', ['cot_id'], ['id'])
        batch_op.create_foreign_key(None, 'points', ['point_id'], ['id'])
        batch_op.create_foreign_key(None, 'euds', ['sender_uid'], ['uid'])

    with op.batch_alter_table('apscheduler_jobs', schema=None) as batch_op:
        batch_op.alter_column('id',
               existing_type=sa.String(length=255),
               type_=sa.VARCHAR(length=191),
               existing_nullable=False)

    with op.batch_alter_table('alerts', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key(None, 'points', ['point_id'], ['id'])
        batch_op.create_foreign_key(None, 'cot', ['cot_id'], ['id'])
        batch_op.create_foreign_key(None, 'euds', ['sender_uid'], ['uid'])

    # ### end Alembic commands ###



import datetime
import traceback

import bleach
import pika
from OpenSSL.crypto import X509
from flask import Blueprint, current_app as app, jsonify, request
from flask_babel import gettext
from flask_security import current_user

from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db, logger, ldap_manager
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser

group_api = Blueprint('group_api', __name__)


@group_api.route('/Marti/api/groups/groupCacheEnabled')
def group_cache_enabled():
    return jsonify(
        {"version": "3", "type": "java.lang.Boolean", "nodeId": app.config.get("OTS_NODE_ID"),
         "data": app.config.get("OTS_ENABLE_CHANNELS")})


@group_api.route('/Marti/api/groups/all')
def get_all_groups():
    cert = verify_client_cert()
    if not cert:
        return jsonify({"success": False, "error": gettext(u"Groups are only supported on SSL connections")}), 400

    response = {"version": "3", "type": "com.bbn.marti.remote.groups.Group", "nodeId": app.config.get("OTS_NODE_ID"), "data": []}

    username = cert.get_subject().commonName

    if not app.config.get("OTS_ENABLE_LDAP"):
        user = app.security.datastore.find_user(username=username)
        groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id)).scalars()

        number_of_groups = 0

        if groups:
            for group in groups:
                if group.direction == Group.IN:
                    response['data'].append(group.group.to_marti_json_in())
                else:
                    response['data'].append(group.group.to_marti_json_out())
                number_of_groups += 1

        # If a user is not assigned to any groups, default them to the __ANON__ group
        if not groups or not number_of_groups:
            logger.info(f"{username} has no groups, defaulting to __ANON__")
            group = Group()
            group.name = "__ANON__"
            group.created = datetime.datetime.now(datetime.timezone.utc)
            group.type = Group.SYSTEM
            group.bitpos = 2

            response['data'].append(group.to_marti_json_out())
            response['data'].append(group.to_marti_json_in())

    else:
        groups = ldap_manager.get_user_groups(username)
        for group in groups:
            if app.config.get("OTS_LDAP_GROUP_PREFIX") and not group['cn'].startswith(app.config.get("OTS_LDAP_GROUP_PREFIX")):
                continue

            g = Group()
            g.name = group['cn']
            g.distinguishedName = group['dn']
            g.type = Group.LDAP

            if group['cn'].lower().endswith("_write"):
                response['data'].append(g.to_marti_json_in())
            elif group['cn'].lower().endswith("_read"):
                response['data'].append(g.to_marti_json_out())

    return jsonify(response)


@group_api.route('/Marti/api/groups')
def get_ldap_groups():

    group_name = request.args.get("groupNameFilter")
    if not group_name:
        return jsonify({'success': False, 'error': gettext(u"Please specify a groupNameFilter")}), 400

    group_name = bleach.clean(group_name)

    if not group_name.startswith(app.config.get("OTS_LDAP_GROUP_PREFIX")):
        return jsonify({'success': False, 'error': gettext(u"Please specify a groupNameFilter that starts with %(prefix)s", prefix=app.config.get('OTS_LDAP_GROUP_PREFIX'))}), 400

    if not group_name.endswith("_READ") and not group_name.endswith("_WRITE"):
        return jsonify({'success': False,'error': gettext(u"groupNameFilter must end with either _READ or _WRITE")}), 400

    response = {"version": "3", "type": "com.bbn.marti.remote.groups.LdapGroup", "data": [],
                "nodeId": app.config.get("OTS_NODE_ID")}

    if app.config.get("OTS_ENABLE_LDAP"):
        group = ldap_manager.get_group_info(group_name)
        g = Group()
        g.name = group['cn']
        g.distinguishedName = group['dn']
        g.type = Group.LDAP

        if group['cn'].endswith("_WRITE"):
            response['data'].append(g.to_marti_json_in())
        elif group['cn'].endswith("_READ"):
            response['data'].append(g.to_marti_json_out())

    return jsonify(response)


@group_api.route('/Marti/api/groups/members')
def get_ldap_group_members():
    group_name = request.args.get("groupNameFilter")
    if not group_name:
        return jsonify({'success': False, 'error': gettext(u"Please specify a groupNameFilter")}), 400

    group_name = bleach.clean(group_name)

    if not group_name.startswith(app.config.get("OTS_LDAP_GROUP_PREFIX")):
        return jsonify({'success': False, 'error': gettext(u"Please specify a groupNameFilter that starts with %(prefix)s", prefix=app.config.get('OTS_LDAP_GROUP_PREFIX'))}), 400

    if not group_name.endswith("_READ") and not group_name.endswith("_WRITE"):
        return jsonify({'success': False, 'error': gettext(u"groupNameFilter must end with either _READ or _WRITE")}), 400

    response = {"version": "3", "type": "java.lang.Integer", "data": 0, "nodeId": app.config.get("OTS_NODE_ID")}

    if app.config.get("OTS_ENABLE_LDAP"):
        group = {}
        try:
            group = ldap_manager.get_group_info(group_name)
            response['data'] = len(group['member'])
        except BaseException as e:
            logger.error(f"Failed to get member count for {group_name}: {e}")
            logger.debug(group)

    return jsonify(response)


@group_api.route('/Marti/api/groupprefix')
def get_ldap_group_prefix():
    response = {"version": "3", "type": "java.lang.String", "data": "", "nodeId": app.config.get("OTS_NODE_ID")}
    if app.config.get("OTS_ENABLE_LDAP"):
        response["data"] = app.config.get("OTS_LDAP_GROUP_PREFIX")

    return jsonify(response)


@group_api.route('/Marti/api/groups/activebits', methods=['PUT'])
def put_active_bits():
    client_uid = request.args.get("clientUid")
    bits = request.json

    return '', 200


@group_api.route('/Marti/api/groups/active', methods=['PUT'])
def put_active_groups():
    logger.info(request.args)
    logger.info(request.json)
    uid = request.args.get("clientUid")
    if not uid:
        logger.error("clientUid required")
        return jsonify({'success': False, 'error': gettext(u"clientUid required")}), 400

    cert = verify_client_cert()
    username = cert.get_subject().commonName
    user = app.security.datastore.find_user(username=username)

    groups_users = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id)).all()
    if not groups_users:
        logger.warning(f"User {username} doesn't belong to any groups, defaulting to __ANON__")
        return '', 400

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()

    group_subscriptions = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id)).all()

    for subscription in request.json:
        direction = subscription.get("direction")
        if not direction and direction != Group.IN and direction != Group.OUT:
            logger.error(f"Direction must be IN or OUT: {direction}")
            return jsonify({"success": False, "error": gettext(u"Direction must be IN or OUT")}), 400

        active = subscription.get("active")
        if not isinstance(active, bool):
            logger.error("The active attribute must be true or false")
            return jsonify({"success": False, "error": gettext(u"The active attribute must be true or false")}), 400

        group_name = subscription.get("name")
        if not group_name:
            logger.error("Group name is required")
            return jsonify({"success": False, "error": gettext(u"Group name is required")}), 400

        group_name = bleach.clean(group_name)

        user_in_group = False
        for group_subscription in group_subscriptions:
            group_subscription = group_subscription[0]
            if group_subscription.group.name == group_name and group_subscription.direction == direction:
                if not app.config.get("OTS_ENABLE_LDAP"):
                    group_subscription.enabled = active
                    db.session.add(group_subscription)

                if active:
                    channel.queue_bind(queue=uid, exchange="groups", routing_key=f"{group_subscription.group.name}.{group_subscription.direction}")
                else:
                    channel.queue_unbind(queue=uid, exchange="groups", routing_key=f"{group_subscription.group.name}.{group_subscription.direction}")

                user_in_group = True

        if not user_in_group:
            logger.warning(f"{username} is not in the {group_name} group")
            db.session.rollback()
            channel.close()
            rabbit_connection.close()
            return jsonify({"success": False, "error": gettext(u"%(username)s is not in the %(group_name)s group", username=username, group_name=group_name)}), 403

    try:
        channel.close()
        rabbit_connection.close()
        db.session.commit()
        return '', 200
    except BaseException as e:
        logger.error(f"Failed to update group subscriptions for {current_user.username}: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": gettext("Failed to update group subscriptions for %(username)s: %(e)s",
                                                           username=current_user.username, e=str(e))}), 400


@group_api.route('/Marti/api/groups/update/<username>')
def update_group(username: str):

    response = {
        "version": "",
        "type": "",
        "data": True,
        "messages": [""],
        "nodeId": app.config.get("OTS_NODE_ID")
    }

    return jsonify(response)


@group_api.route('/Marti/api/groups/<group_name>/<direction>')
def get_group(group_name: str, direction: str):
    if not group_name or not direction:
        return jsonify({'success': False, 'error': gettext("Please provide a group name and direction")}), 400

    response = {"version": "3", "type": "com.bbn.marti.remote.groups.Group", "data": {}, "nodeId": app.config.get("OTS_NODE_ID")}

    if not app.config.get("OTS_ENABLE_LDAP"):
        group = db.session.execute(db.session.query(Group).filter_by(group_name=group_name, direction=direction)).first()
        if not group:
            return jsonify({'success': False, 'error': gettext("No group found: %(group_name)s, %(direction)s", group_name=group_name, direction=direction)}), 404
        if direction == Group.IN:
            response['data'] = group[0].to_marti_json_in()
        elif direction == Group.OUT:
            response['data'] = group[0].to_marti_json_out()

    else:
        group = ldap_manager.get_group_info(group_name)
        g = Group()
        g.name = group['cn']
        g.distinguishedName = group['dn']
        g.type = Group.LDAP
        if direction == Group.IN:
            response['data'] = g.to_marti_json_in()
        elif direction == Group.OUT:
            response['data'] = g.to_marti_json_out()

    return jsonify(response)


@group_api.route('/Marti/api/subscriptions/all')
def get_all_subscriptions():
    sortBy = request.args.get("sortBy")
    direction = request.args.get("direction")
    page = request.args.get("page")
    limit = request.args.get("limit")

    response = {"version": "3", "type": "SubscriptionInfo", "data": [], "messages": [], "nodeId": app.config.get("OTS_NODE_ID")}

    return jsonify(response)


@group_api.route('/Marti/api/groups/activeForce')
def group_active_force():
    username = request.args.get("username")
    if not username:
        return '', 400

    response = {
        "name": "",
        "distinguishedName": "",
        "direction": "",
        "created": "",
        "type": "SYSTEM",  # Can also be LDAP but OTS doesn't support LDAP yet
        "bitpos": 0,
        "active": True,
        "description": ""
    }

    return jsonify(response)


@group_api.route('/Marti/api/groups/user')
def get_user_groups():
    username = request.args.get("username")
    if not username:
        return '', 400

    groups = {}

    response = {
        "version": "",
        "type": "",
        "data": groups,
        "messages": [""],
        "nodeId": app.config.get("OTS_NODE_ID")
    }

    return jsonify(response)



"""Added mission_name to cot table

Revision ID: d31bff4a15c7
Revises: 5d06227dea50
Create Date: 2024-10-09 15:40:49.385642

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd31bff4a15c7'
down_revision = '5d06227dea50'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.add_column(sa.Column('mission_name', sa.String(length=255), nullable=True))
        batch_op.create_foreign_key("mission_name", 'missions', ['mission_name'], ['name'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.drop_constraint("mission_name", type_='foreignkey')
        batch_op.drop_column('mission_name')

    # ### end Alembic commands ###



import json
import math
import re
from datetime import datetime, timezone
from xml.etree.ElementTree import Element, SubElement, tostring

from flask import current_app as app
import pika.channel

ISO8601_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
ISO8601_FORMAT_NO_MICROSECONDS = "%Y-%m-%dT%H:%M:%SZ"
affiliations = ['friendly', 'hostile', 'unknown', 'pending', 'assumed', 'neutral', 'suspect', 'joker', 'faker']

# For WTForms BooleanField, the default doesn't include 'False'
# https://wtforms.readthedocs.io/en/3.1.x/fields/?highlight=false_values#wtforms.fields.BooleanField
false_values = (False, 'False', 'false', '')


def get_tasking(cot_type):
    if re.match("^t-x-f", cot_type):
        return "remarks"
    if re.match("^t-x-s", cot_type):
        return "state/sync"
    if re.match("^t-s", cot_type):
        return "required"
    if re.match("^t-z", cot_type):
        return "cancel"
    if re.match("^t-x-c-c", cot_type):
        return "commcheck"
    if re.match("^t-x-c-g-d", cot_type):
        return "dgps"
    if re.match("^t-k-d", cot_type):
        return "destroy"
    if re.match("^t-k-i", cot_type):
        return "investigate"
    if re.match("^t-k-t", cot_type):
        return "target"
    if re.match("^t-k", cot_type):
        return "strike"
    if re.match("^t-", cot_type):
        return "tasking"
    return None


def get_affiliation(cot_type):
    if re.match("^t-", cot_type):
        return get_tasking(cot_type)
    if re.match("^a-f-", cot_type):
        return "friendly"
    if re.match("^a-h-", cot_type):
        return "hostile"
    if re.match("^a-u-", cot_type):
        return "unknown"
    if re.match("^a-p-", cot_type):
        return "pending"
    if re.match("^a-a-", cot_type):
        return "assumed"
    if re.match("^a-n-", cot_type):
        return "neutral"
    if re.match("^a-s-", cot_type):
        return "suspect"
    if re.match("^a-j-", cot_type):
        return "joker"
    if re.match("^a-k-", cot_type):
        return "faker"
    return None


def get_battle_dimension(cot_type):
    if re.match("^a-.-A", cot_type):
        return "airborne"
    if re.match("^a-.-G", cot_type):
        return "ground"
    if re.match("^a-.-G-I", cot_type):
        return "installation"
    if re.match("^a-.-S", cot_type):
        return "surface/sea"
    if re.match("^a-.-U", cot_type):
        return "subsurface"
    if re.match("^a-.-F", cot_type):
        return "sof"
    if re.match("^a-.-Z", cot_type):
        return "unknown"
    if re.match("^a-.-P", cot_type):
        return "space"
    return None


def parse_type(cot_type):
    if re.match("^a-.-G-I", cot_type):
        return "installation"
    if re.match("^a-.-G-E-V", cot_type):
        return "vehicle"
    if re.match("^a-.-G-E", cot_type):
        return "equipment"
    if re.match("^a-.-A-W-M-S", cot_type):
        return "sam"
    if re.match("^a-.-A-M-F-Q-r", cot_type):
        return "uav"


def cot_type_to_2525c(cot_type):
    mil_std_2525c = "s"
    cot_type_list = cot_type.split("-")
    cot_type_list.pop(0)  # this should always be letter a
    affiliation = cot_type_list.pop(0)
    battle_dimension = cot_type_list.pop(0)
    mil_std_2525c += affiliation
    mil_std_2525c += battle_dimension
    mil_std_2525c += "-"

    for letter in cot_type_list:
        if letter.isupper():
            mil_std_2525c += letter.lower()

    while len(mil_std_2525c) < 10:
        mil_std_2525c += "-"

    return mil_std_2525c


def datetime_from_iso8601_string(datetime_string: str | None) -> datetime:
    if not datetime_string:
        return datetime.now(timezone.utc)
    try:
        dt = datetime.strptime(datetime_string, ISO8601_FORMAT)
        dt = dt.replace(tzinfo=timezone.utc)
        return dt
    except ValueError:
        dt = datetime.strptime(datetime_string, ISO8601_FORMAT_NO_MICROSECONDS)
        dt = dt.replace(tzinfo=timezone.utc)
        return dt


def iso8601_string_from_datetime(datetime_object):
    if datetime_object:
        return datetime_object.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-2] + "Z"
    else:
        return None


def iso8601_string_from_datetime_no_ms(datetime_object):
    if datetime_object:
        return datetime_object.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        return None


def generate_delete_cot(uid: str, cot_type: str) -> Element:
    now = datetime.now(timezone.utc)

    event = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-d-d', 'version': '2.0',
                              'uid': uid, 'start': iso8601_string_from_datetime(now),
                              'time': iso8601_string_from_datetime(now),
                              'stale': iso8601_string_from_datetime(now)})
    SubElement(event, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                'lon': '0'})
    detail = SubElement(event, 'detail')
    SubElement(detail, 'link', {'relation': 'p-p', 'uid': uid, 'type': cot_type})
    SubElement(detail, '_flow-tags_',
               {'TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0': iso8601_string_from_datetime(now)})

    return event


def publish_cot(cot: Element, channel: pika.channel.Channel):
    channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
        {'cot': tostring(cot).decode('utf-8'), 'uid': app.config['OTS_NODE_ID']}),
                          properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))


def format_bytes(size_bytes: int | None):
    if size_bytes is None or size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


def bytes_to_megabytes(size_bytes: int | None):
    if not size_bytes:
        return 0
    return round(size_bytes / 1024**2, 2)


def bytes_to_gigabytes(size_bytes: int | None):
    if not size_bytes:
        return 0
    return round(size_bytes / 1024**3, 2)



from dataclasses import dataclass
from datetime import datetime

from opentakserver.extensions import db

# Leave these imports, they're needed when making new DB migrations
from opentakserver.models.Team import Team
from opentakserver.models.Chatrooms import Chatroom

from sqlalchemy import Integer, String, ForeignKey, DateTime, BigInteger
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class EUD(db.Model):
    __tablename__ = "euds"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    device: Mapped[str] = mapped_column(String(255), nullable=True)
    os: Mapped[str] = mapped_column(String(255), nullable=True)
    platform: Mapped[str] = mapped_column(String(255), nullable=True)
    version: Mapped[str] = mapped_column(String(255), nullable=True)
    phone_number: Mapped[int] = mapped_column(BigInteger, nullable=True)
    last_event_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    last_status: Mapped[str] = mapped_column(String(255), nullable=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), nullable=True)
    team_id: Mapped[int] = mapped_column(Integer, ForeignKey("teams.id"), nullable=True)
    team_role: Mapped[str] = mapped_column(String(255), nullable=True)
    meshtastic_id: Mapped[int] = mapped_column(BigInteger, nullable=True)
    meshtastic_macaddr: Mapped[String] = mapped_column(String(255), nullable=True)
    last_meshtastic_publish: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    points = relationship("Point", cascade="all, delete-orphan", back_populates="eud")
    cots = relationship("CoT", cascade="all, delete-orphan", back_populates="eud")
    casevacs = relationship("CasEvac", cascade="all, delete-orphan", back_populates="eud")
    geochats = relationship("GeoChat", cascade="all, delete-orphan", back_populates="eud")
    chatroom_uid = relationship("ChatroomsUids", cascade="all, delete-orphan", back_populates="eud")
    user = relationship("User", back_populates="euds")
    alert = relationship("Alert", cascade="all, delete-orphan", back_populates="eud")
    data_packages = relationship("DataPackage", cascade="all, delete-orphan", back_populates="eud", uselist=False)
    certificate = relationship("Certificate", cascade="all, delete, delete-orphan", back_populates="eud", uselist=False)
    markers = relationship("Marker", cascade="all, delete, delete-orphan", back_populates="eud")
    rb_lines = relationship("RBLine", cascade="all, delete-orphan", back_populates="eud")
    team = relationship("Team", back_populates="euds")
    owned_missions = relationship("Mission", back_populates="owner")
    stats = relationship("EUDStats", back_populates="eud")

    def serialize(self):
        return {
            'uid': self.uid,
            'callsign': self.callsign,
            'device': self.device,
            'os': self.os,
            'platform': self.platform,
            'version': self.version,
            'phone_number': self.phone_number,
            'last_event_time': self.last_event_time,
            'last_status': self.last_status,
            'user_id': self.user_id,
            'team_id': self.team_id,
            'team_role': self.team_role
        }

    def to_json(self, include_data_packages=True):
        config_datapackage_hash = None
        if self.certificate and self.certificate.data_package:
            config_datapackage_hash = self.certificate.data_package.hash
        return {
            'uid': self.uid,
            'callsign': self.callsign,
            'device': self.device,
            'os': self.os,
            'platform': self.platform,
            'version': self.version,
            'phone_number': self.phone_number,
            'last_event_time': iso8601_string_from_datetime(self.last_event_time) if self.last_event_time else None,
            'last_status': self.last_status,
            'username': self.user.username if self.user else None,
            'last_point': self.points[-1].to_json() if self.points else None,
            'team': self.team.name if self.team else None,
            'team_color': self.team.get_team_color() if self.team else None,
            'team_role': self.team_role,
            'data_packages': self.data_packages.to_json(False) if include_data_packages and self.data_packages else None,
            'config_hash': config_datapackage_hash
        }



import datetime
from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class MissionInvitation(db.Model):
    __tablename__ = "mission_invitations"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey('missions.name'))
    client_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True)
    callsign: Mapped[str] = mapped_column(String(255), ForeignKey("euds.callsign", ondelete="CASCADE"), nullable=True)
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True)
    group_name: Mapped[str] = mapped_column(String(255), nullable=True)
    team_name: Mapped[str] = mapped_column(String(255), ForeignKey("teams.name"), nullable=True)
    creator_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    role: Mapped[str] = mapped_column(String(255), nullable=True)

    eud_uid = relationship("EUD", foreign_keys=[client_uid], uselist=False)
    eud_callsign = relationship("EUD", foreign_keys=[callsign], uselist=False)
    user = relationship("User", back_populates="mission_invitations", uselist=False)
    team = relationship("Team", back_populates="mission_invitations", uselist=False)
    mission = relationship("Mission", back_populates="invitations", uselist=False)

    def serialize(self):
        return {
            'mission_name': self.mission_name,
            'client_uid': self.client_uid,
            'callsign': self.callsign,
            'username': self.username,
            'group_name': self.group,
            'team_name': self.team_name,
            'creator_uid': self.creator_uid,
            'role': self.role
        }

    def to_json(self):
        return self.serialize()



from opentakserver.extensions import db
from sqlalchemy import String, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


class ChatroomsUids(db.Model):
    __tablename__ = 'chatrooms_uids'

    chatroom_id: Mapped[str] = mapped_column(String(255), ForeignKey("chatrooms.id"), primary_key=True)
    uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), primary_key=True)
    chatroom = relationship("Chatroom", back_populates="chatroom_uid")
    eud = relationship("EUD", back_populates="chatroom_uid")



from dataclasses import dataclass

from opentakserver.extensions import db
from opentakserver.functions import bytes_to_gigabytes, bytes_to_megabytes
from sqlalchemy import Integer, String, ForeignKey, DateTime, BigInteger
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class EUDStats(db.Model):
    __tablename__ = "eud_stats"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    timestamp: Mapped[DateTime] = mapped_column(DateTime, nullable=True)
    eud_uid: Mapped[str] = mapped_column(String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=False)
    heap_free_size: Mapped[int] = mapped_column(BigInteger, nullable=True)
    app_framerate: Mapped[int] = mapped_column(Integer, nullable=True)
    storage_total: Mapped[int] = mapped_column(BigInteger, nullable=True)
    heap_current_size: Mapped[int] = mapped_column(Integer, nullable=True)
    battery: Mapped[int] = mapped_column(Integer, nullable=True)
    battery_temp: Mapped[int] = mapped_column(Integer, nullable=True)
    deviceDataRx: Mapped[int] = mapped_column(BigInteger, nullable=True)
    heap_max_size: Mapped[int] = mapped_column(BigInteger, nullable=True)
    storage_available: Mapped[int] = mapped_column(BigInteger, nullable=True)
    deviceDataTx: Mapped[int] = mapped_column(BigInteger, nullable=True)
    ip_address: Mapped[str] = mapped_column(String(255), nullable=True)
    battery_status: Mapped[str] = mapped_column(String(255), nullable=True)
    eud = relationship("EUD", back_populates="stats", uselist=False)

    def serialize(self):
        return {
            "timestamp": self.timestamp,
            "eud_uid": self.eud_uid,
            "heap_free_size": self.heap_free_size,
            "app_framerate": self.app_framerate,
            "storage_total": self.storage_total,
            "heap_current_size": self.heap_current_size,
            "battery": self.battery,
            "deviceDataRx": self.deviceDataRx,
            "heap_max_size": self.heap_max_size,
            "storage_available": self.storage_available,
            "deviceDataTx": self.deviceDataTx,
            "ip_address": self.ip_address,
            "battery_status": self.battery_status
        }

    def to_json(self):
        return {
            "timestamp": iso8601_string_from_datetime(self.timestamp),
            "eud_uid": self.eud_uid,
            "heap_free_size": bytes_to_megabytes(self.heap_free_size),
            "app_framerate": self.app_framerate,
            "storage_total": bytes_to_gigabytes(self.storage_total),
            "heap_current_size": bytes_to_megabytes(self.heap_current_size),
            "battery": self.battery,
            "deviceDataRx": bytes_to_megabytes(self.deviceDataRx),
            "heap_max_size": bytes_to_megabytes(self.heap_max_size),
            "storage_available": bytes_to_gigabytes(self.storage_available),
            "deviceDataTx": bytes_to_megabytes(self.deviceDataTx),
            "ip_address": self.ip_address,
            "battery_status": self.battery_status
        }



"""Updated MissionLogEntry

Revision ID: 35a075531973
Revises: 2b1863a985a6
Create Date: 2025-11-12 05:01:51.094540

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '35a075531973'
down_revision = '2b1863a985a6'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('mission_logs', schema=None) as batch_op:
        batch_op.add_column(sa.Column('content_hashes', sa.JSON(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_logs', schema=None) as batch_op:
        batch_op.drop_column('content_hashes')

    # ### end Alembic commands ###
